#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

from kiwi.cli.better_argparse import HyperPipelineParser
from kiwi.cli.opts import PathType
from kiwi.cli.pipelines import train
from kiwi.lib import jackknife

logger = logging.getLogger(__name__)


def jackknife_opts(parser):
    # Training loop options
    group = parser.add_argument_group('jackknifing')

    group.add(
        '--splits',
        required=False,
        type=int,
        default=5,
        help='Jackknife with X folds.',
    )
    group.add(
        '--train-config',
        required=False,
        type=PathType(exists=True),
        help='Path to config file with model parameters.',
    )


def build_parser():
    return HyperPipelineParser(
        name='jackknife',
        pipeline_parser=train.build_parser(),
        pipeline_config_key='train-config',
        options_fn=jackknife_opts,
    )


def main(argv=None):
    parser = build_parser()
    options = parser.parse(args=argv)
    jackknife.run_from_options(options)


if __name__ == '__main__':  # pragma: no cover
    main()  # pragma: no cover



"""This is the main script for the linear sequential word-based quality
estimator."""
#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from pathlib import Path

import numpy as np

from kiwi import constants as const
from kiwi.data.fieldsets.linear import build_fieldset
from kiwi.models.linear.label_dictionary import LabelDictionary
from kiwi.models.linear.linear_word_qe_decoder import LinearWordQEDecoder
from kiwi.models.linear.linear_word_qe_features import LinearWordQEFeatures
from kiwi.models.linear.linear_word_qe_sentence import LinearWordQESentence
from kiwi.models.linear.sequence_parts import (
    SequenceBigramPart,
    SequenceUnigramPart,
)
from kiwi.models.linear.structured_classifier import StructuredClassifier

logger = logging.getLogger(__name__)


class LinearWordQEClassifier(StructuredClassifier):
    """Main class for the word-level quality estimator. Inherits from a
    general structured classifier."""

    title = 'Linear Model'

    def __init__(
        self,
        use_basic_features_only=True,
        use_bigrams=True,
        use_simple_bigram_features=True,
        use_parse_features=False,
        use_stacked_features=False,
        evaluation_metric='f1_bad',
        cost_false_positives=0.5,
        cost_false_negatives=0.5,
    ):
        super().__init__()

        self.decoder = LinearWordQEDecoder(
            self, cost_false_positives, cost_false_negatives
        )
        self.labels = LabelDictionary()

        self.use_basic_features_only = use_basic_features_only
        self.use_bigrams = use_bigrams
        self.use_simple_bigram_features = use_simple_bigram_features
        self.use_parse_features = use_parse_features
        self.use_stacked_features = use_stacked_features

        # Evaluation.
        self.evaluation_metric = evaluation_metric

    @staticmethod
    def fieldset(*args, **kwargs):
        return build_fieldset()

    @staticmethod
    def from_options(vocabs, opts):
        use_parse_features = True if opts.train_target_parse else False
        use_stacked_features = True if opts.train_target_stacked else False
        model = LinearWordQEClassifier(
            use_basic_features_only=opts.use_basic_features_only,
            use_bigrams=opts.use_bigrams,
            use_simple_bigram_features=opts.use_simple_bigram_features,
            use_parse_features=use_parse_features,
            use_stacked_features=use_stacked_features,
            evaluation_metric=opts.evaluation_metric,
            cost_false_positives=opts.cost_false_positives,
            cost_false_negatives=opts.cost_false_negatives,
        )
        return model

    def num_parameters(self):
        return len(self.__dict__)

    # -- END of new methods --

    # TODO: Eliminate this function.
    def get_coarse_label(self, label):
        """Get the coarse part of a fine-grained label. The coarse label is the
        prefix before the underscore (if any). For example, the coarse part of
        BAD_SUB, BAD_DEL, and BAD is BAD."""
        sep = label.find('_')
        if sep >= 0:
            coarse_label = label[:sep]
        else:
            coarse_label = label
        return coarse_label

    def create_instances(self, dataset):
        instances = []
        num_words = 0
        for example in dataset:
            sentence = LinearWordQESentence()
            labels = None
            if hasattr(example, 'tags'):
                labels = []
                for label in example.tags:
                    if label in self.labels:
                        label_id = self.labels.get_label_id(label)
                    else:
                        label_id = self.labels.add(label)
                    labels.append(label_id)

            sentence.create_from_sentence_pair(
                source_words=example.source,
                target_words=example.target,
                alignments=example.alignments,
                source_pos_tags=getattr(example, const.SOURCE_POS, None),
                target_pos_tags=getattr(example, const.TARGET_POS, None),
                target_parse_heads=getattr(
                    example, const.TARGET_PARSE_HEADS, None
                ),
                target_parse_relations=getattr(
                    example, const.TARGET_PARSE_RELATIONS, None
                ),
                target_ngram_left=getattr(
                    example, const.TARGET_NGRAM_LEFT, None
                ),
                target_ngram_right=getattr(
                    example, const.TARGET_NGRAM_RIGHT, None
                ),
                target_stacked_features=getattr(
                    example, const.TARGET_STACKED, None
                ),
                labels=labels,
            )
            instances.append(sentence)
            num_words += sentence.num_words()

        logger.info('Number of sentences: %d' % len(instances))
        logger.info('Number of words: %d' % num_words)
        logger.info('Number of labels: %d' % len(self.labels))

        return instances

    def make_parts(self, instance):
        """Creates the parts (unigrams and bigrams) for an instance."""
        gold_list = []
        parts = []
        make_gold = True
        for word_index in range(instance.num_words()):
            for label_id in range(len(self.labels)):
                part = SequenceUnigramPart(word_index, label_id)
                parts.append(part)
                if make_gold:
                    if label_id == instance.sentence_word_labels[word_index]:
                        gold_list.append(1.0)
                    else:
                        gold_list.append(0.0)
        if self.use_bigrams:
            # First word.
            for label_id in range(len(self.labels)):
                part = SequenceBigramPart(0, label_id, -1)
                parts.append(part)
                if make_gold:
                    if label_id == instance.sentence_word_labels[0]:
                        gold_list.append(1.0)
                    else:
                        gold_list.append(0.0)
            # Intermediate word.
            for word_index in range(1, instance.num_words()):
                for label_id in range(len(self.labels)):
                    for previous_label_id in range(len(self.labels)):
                        part = SequenceBigramPart(
                            word_index, label_id, previous_label_id
                        )
                        parts.append(part)
                        if make_gold:
                            if (
                                label_id
                                == instance.sentence_word_labels[word_index]
                                and previous_label_id
                                == instance.sentence_word_labels[word_index - 1]
                            ):
                                gold_list.append(1.0)
                            else:
                                gold_list.append(0.0)
            # Last word.
            for previous_label_id in range(len(self.labels)):
                part = SequenceBigramPart(
                    instance.num_words(), -1, previous_label_id
                )
                parts.append(part)
                if make_gold:
                    if (
                        previous_label_id
                        == instance.sentence_word_labels[
                            instance.num_words() - 1
                        ]
                    ):
                        gold_list.append(1.0)
                    else:
                        gold_list.append(0.0)

        if make_gold:
            gold_array = np.array(gold_list)
            return parts, gold_array
        else:
            return parts

    def make_features(self, instance, parts):
        """Creates a feature vector for each part."""
        features = []
        for part in parts:
            part_features = LinearWordQEFeatures(
                use_basic_features_only=self.use_basic_features_only,
                use_simple_bigram_features=self.use_simple_bigram_features,
                use_parse_features=self.use_parse_features,
                use_stacked_features=self.use_stacked_features,
            )
            if isinstance(part, SequenceUnigramPart):
                part_features.compute_unigram_features(
                    instance.sentence_word_features, part
                )
            elif isinstance(part, SequenceBigramPart):
                part_features.compute_bigram_features(
                    instance.sentence_word_features, part
                )
            else:
                raise NotImplementedError
            features.append(part_features)
        return features

    def label_instance(self, instance, parts, predicted_output):
        """Return a labeled instance by adding the predicted output
        information."""
        assert False, 'This does not seem to be called'
        labeled_instance = LinearWordQESentence(instance.sentence)
        labeled_instance.sentence_word_features = (
            instance.sentence_word_features
        )
        predictions = np.zeros(instance.num_words(), dtype=int)
        for r, part in enumerate(parts):
            if isinstance(part, SequenceUnigramPart):
                continue
            if predicted_output[r] > 0.5:
                predictions[part.index] = part.label
        labeled_instance.sentence_word_labels = [
            self.labels.get_label_name(pred) for pred in predictions
        ]
        return labeled_instance

    def create_prediction(self, instance, parts, predicted_output):
        """Creates a list of word-level predictions for a sentence.
        For compliance with probabilities, it returns 1 if label is BAD, 0 if
        OK."""
        predictions = np.zeros(instance.num_words(), dtype=int)
        for r, part in enumerate(parts):
            if not isinstance(part, SequenceUnigramPart):
                continue
            if predicted_output[r] > 0.5:
                predictions[part.index] = part.label
        predictions = [
            int(const.BAD == self.labels.get_label_name(pred))
            for pred in predictions
        ]
        return predictions

    def test(self, instances):
        """Run the model on test data."""
        logger.info('Testing...')
        predictions = StructuredClassifier.test(self, instances)
        return predictions

    def evaluate(self, instances, predictions, print_scores=True):
        """Evaluates the model's accuracy and F1-BAD score."""
        all_predictions = []
        for word_predictions in predictions:
            labels = [
                const.BAD if prediction else const.OK
                for prediction in word_predictions
            ]
            labels = [int(self.labels[label]) for label in labels]
            all_predictions.append(labels)

        # TODO: Get rid of fine-grained labels.
        # Allow fine-grained labels. Their names should be a coarse-grained
        # label, followed by an underscore, followed by a sub-label.
        # For example, BAD_SUB or BAD_DEL are two instances of bad labels.
        fine_to_coarse = -np.ones(len(self.labels), dtype=int)
        coarse_labels = LabelDictionary()
        for label in self.labels:
            coarse_label = self.get_coarse_label(label)
            if coarse_label not in coarse_labels:
                lid = coarse_labels.add(coarse_label)
            else:
                lid = coarse_labels[coarse_label]
            fine_to_coarse[self.labels[label]] = lid

        # Iterate through sentences and compare gold values with predicted
        # values. Update counts.
        num_matched = 0
        num_matched_labels = np.zeros(len(coarse_labels))
        num_predicted = 0
        num_predicted_labels = np.zeros(len(coarse_labels))
        num_gold_labels = np.zeros(len(coarse_labels))
        assert len(all_predictions) == len(instances)
        for i, instance in enumerate(instances):
            predictions = all_predictions[i]
            assert len(instance.sentence_word_labels) == len(predictions)
            for j in range(len(predictions)):
                if (
                    fine_to_coarse[predictions[j]]
                    == fine_to_coarse[instance.sentence_word_labels[j]]
                ):
                    num_matched += 1
                num_predicted += 1
                if (
                    fine_to_coarse[predictions[j]]
                    == fine_to_coarse[instance.sentence_word_labels[j]]
                ):
                    num_matched_labels[fine_to_coarse[predictions[j]]] += 1
                num_predicted_labels[fine_to_coarse[predictions[j]]] += 1
                num_gold_labels[
                    fine_to_coarse[instance.sentence_word_labels[j]]
                ] += 1

        acc = float(num_matched) / float(num_predicted)
        logger.info('Accuracy: %f' % acc)

        # We allow multiple bad labels. They should be named BAD*.
        bad = coarse_labels['BAD']
        if num_matched_labels[bad] == 0:
            f1_bad = 0.0
        else:
            precision_bad = float(num_matched_labels[bad]) / float(
                num_predicted_labels[bad]
            )
            recall_bad = float(num_matched_labels[bad]) / float(
                num_gold_labels[bad]
            )
            f1_bad = (
                2 * precision_bad * recall_bad / (precision_bad + recall_bad)
            )

        logger.info(
            '# gold bad: %d/%d' % (num_gold_labels[bad], sum(num_gold_labels))
        )
        logger.info(
            '# predicted bad: %d/%d'
            % (num_predicted_labels[bad], sum(num_predicted_labels))
        )

        ok = coarse_labels['OK']
        if num_matched_labels[ok] == 0:
            f1_ok = 0.0
        else:
            precision_ok = float(num_matched_labels[ok]) / float(
                num_predicted_labels[ok]
            )
            recall_ok = float(num_matched_labels[ok]) / float(
                num_gold_labels[ok]
            )
            f1_ok = 2 * precision_ok * recall_ok / (precision_ok + recall_ok)

        logger.info(
            '# gold ok: %d/%d' % (num_gold_labels[ok], sum(num_gold_labels))
        )
        logger.info(
            '# predicted ok: %d/%d'
            % (num_predicted_labels[ok], sum(num_predicted_labels))
        )

        logger.info('F1 bad: %f' % f1_bad)
        logger.info('F1 ok: %f' % f1_ok)
        logger.info('F1 mult: %f' % (f1_bad * f1_ok))
        if self.evaluation_metric == 'f1_mult':
            return f1_bad * f1_ok
        elif self.evaluation_metric == 'f1_bad':
            return f1_bad
        else:
            raise NotImplementedError

    def load_configuration(self, config):
        self.use_basic_features_only = config['use_basic_features_only']
        self.use_bigrams = config['use_bigrams']
        self.use_simple_bigram_features = config['use_simple_bigram_features']
        self.use_stacked_features = config['use_stacked']
        self.use_parse_features = config['use_parse']

    def save_configuration(self):
        config = {
            'use_basic_features_only': self.use_basic_features_only,
            'use_bigrams': self.use_bigrams,
            'use_simple_bigram_features': self.use_simple_bigram_features,
            'use_stacked': self.use_stacked_features,
            'use_parse': self.use_parse_features,
        }
        return config

    def load(self, model_path):
        import pickle

        with Path(model_path).open('rb') as fid:
            config = pickle.load(fid)
            self.load_configuration(config)
            self.labels = pickle.load(fid)
            self.model = pickle.load(fid)
            try:
                self.source_vocab = pickle.load(fid)
                self.target_vocab = pickle.load(fid)
            except EOFError:
                self.source_vocab = None
                self.target_vocab = None

    def save(self, model_path):
        import pickle

        with Path(model_path).open('wb') as fid:
            config = self.save_configuration()
            pickle.dump(config, fid)
            pickle.dump(self.labels, fid)
            pickle.dump(self.model, fid)
            # pickle.dump(self.source_vocab, fid)
            # pickle.dump(self.target_vocab, fid)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F

from kiwi import constants as const
from kiwi.data.fieldsets.quetch import build_fieldset
from kiwi.models.model import Model
from kiwi.models.quetch import QUETCH
from kiwi.models.utils import make_loss_weights


@Model.register_subclass
class NuQE(QUETCH):
    """Neural Quality Estimation (NuQE) model for word level quality
    estimation."""

    title = 'NuQE'

    def __init__(self, vocabs, **kwargs):

        self.source_emb = None
        self.target_emb = None
        self.linear_1 = None
        self.linear_2 = None
        self.linear_3 = None
        self.linear_4 = None
        self.linear_5 = None
        self.linear_6 = None
        self.linear_out = None
        self.embeddings_dropout = None
        self.dropout = None
        self.gru1 = None
        self.gru2 = None
        self.is_built = False
        super().__init__(vocabs, **kwargs)

    def build(self, source_vectors=None, target_vectors=None):
        nb_classes = self.config.nb_classes
        # FIXME: Remove dependency on magic number
        weight = make_loss_weights(
            nb_classes, const.BAD_ID, self.config.bad_weight
        )

        self._loss = nn.CrossEntropyLoss(
            weight=weight, ignore_index=self.config.tags_pad_id, reduction='sum'
        )

        # Embeddings layers:
        self._build_embeddings(source_vectors, target_vectors)

        feature_set_size = (
            self.config.source_embeddings_size
            + self.config.target_embeddings_size
        ) * self.config.window_size

        l1_dim = self.config.hidden_sizes[0]
        l2_dim = self.config.hidden_sizes[1]
        l3_dim = self.config.hidden_sizes[2]
        l4_dim = self.config.hidden_sizes[3]

        nb_classes = self.config.nb_classes
        dropout = self.config.dropout

        # Linear layers
        self.linear_1 = nn.Linear(feature_set_size, l1_dim)
        self.linear_2 = nn.Linear(l1_dim, l1_dim)
        self.linear_3 = nn.Linear(2 * l2_dim, l2_dim)
        self.linear_4 = nn.Linear(l2_dim, l2_dim)
        self.linear_5 = nn.Linear(2 * l2_dim, l3_dim)
        self.linear_6 = nn.Linear(l3_dim, l4_dim)

        # Output layer
        self.linear_out = nn.Linear(l4_dim, nb_classes)

        # Recurrent Layers
        self.gru_1 = nn.GRU(
            l1_dim, l2_dim, bidirectional=True, batch_first=True
        )
        self.gru_2 = nn.GRU(
            l2_dim, l2_dim, bidirectional=True, batch_first=True
        )

        # Dropout after linear layers
        self.dropout_in = nn.Dropout(dropout)
        self.dropout_out = nn.Dropout(dropout)

        # Explicit initializations
        nn.init.xavier_uniform_(self.linear_1.weight)
        nn.init.xavier_uniform_(self.linear_2.weight)
        nn.init.xavier_uniform_(self.linear_3.weight)
        nn.init.xavier_uniform_(self.linear_4.weight)
        nn.init.xavier_uniform_(self.linear_5.weight)
        nn.init.xavier_uniform_(self.linear_6.weight)
        # nn.init.xavier_uniform_(self.linear_out)
        nn.init.constant_(self.linear_1.bias, 0.0)
        nn.init.constant_(self.linear_2.bias, 0.0)
        nn.init.constant_(self.linear_3.bias, 0.0)
        nn.init.constant_(self.linear_4.bias, 0.0)
        nn.init.constant_(self.linear_5.bias, 0.0)
        nn.init.constant_(self.linear_6.bias, 0.0)
        # nn.init.constant_(self.linear_out.bias, 0.)

        self.is_built = True

    @staticmethod
    def fieldset(*args, **kwargs):
        return build_fieldset(*args, **kwargs)

    @staticmethod
    def from_options(vocabs, opts):
        model = NuQE(
            vocabs=vocabs,
            predict_target=opts.predict_target,
            predict_gaps=opts.predict_gaps,
            predict_source=opts.predict_source,
            source_embeddings_size=opts.source_embeddings_size,
            target_embeddings_size=opts.target_embeddings_size,
            hidden_sizes=opts.hidden_sizes,
            bad_weight=opts.bad_weight,
            window_size=opts.window_size,
            max_aligned=opts.max_aligned,
            dropout=opts.dropout,
            embeddings_dropout=opts.embeddings_dropout,
            freeze_embeddings=opts.freeze_embeddings,
        )
        return model

    def forward(self, batch):
        assert self.is_built

        if self.config.predict_source:
            align_side = const.SOURCE_TAGS
        else:
            align_side = const.TARGET_TAGS

        target_input, source_input, nb_alignments = self.make_input(
            batch, align_side
        )

        #
        # Source Branch
        #
        # (bs, ts, aligned, window) -> (bs, ts, aligned, window, emb)
        h_source = self.source_emb(source_input)
        h_source = self.embeddings_dropout(h_source)

        if len(h_source.shape) == 5:
            # (bs, ts, aligned, window, emb) -> (bs, ts, window, emb)
            h_source = h_source.sum(2, keepdim=False) / nb_alignments.unsqueeze(
                -1
            ).unsqueeze(-1)

        # (bs, ts, window, emb) -> (bs, ts, window * emb)
        h_source = h_source.view(source_input.size(0), source_input.size(1), -1)

        #
        # Target Branch
        #
        # (bs, ts * window) -> (bs, ts * window, emb)
        h_target = self.target_emb(target_input)
        h_target = self.embeddings_dropout(h_target)

        if len(h_target.shape) == 5:
            # (bs, ts, aligned, window, emb) -> (bs, ts, window, emb)
            h_target = h_target.sum(2, keepdim=False) / nb_alignments.unsqueeze(
                -1
            ).unsqueeze(-1)

        # (bs, ts * window, emb) -> (bs, ts, window * emb)
        h_target = h_target.view(target_input.size(0), target_input.size(1), -1)

        #
        # POS tags branches
        #
        feature_set = (h_source, h_target)

        #
        # Merge Branches
        #
        # (bs, ts, window * emb) -> (bs, ts, 2 * window * emb)
        h = torch.cat(feature_set, dim=-1)
        h = self.dropout_in(h)

        #
        # First linears
        #
        # (bs, ts, 2 * window * emb) -> (bs, ts, l1_dim)
        h = F.relu(self.linear_1(h))

        # (bs, ts, l1_dim) -> (bs, ts, l1_dim)
        h = F.relu(self.linear_2(h))

        #
        # First recurrent
        #
        # (bs, ts, l1_dim) -> (bs, ts, l1_dim)
        h, _ = self.gru_1(h)

        #
        # Second linears
        #
        # (bs, ts, l1_dim) -> (bs, ts, l2_dim)
        h = F.relu(self.linear_3(h))

        # (bs, ts, l2_dim) -> (bs, ts, l2_dim)
        h = F.relu(self.linear_4(h))

        #
        # Second recurrent
        #
        # (bs, ts, l2_dim) -> (bs, ts, l2_dim)
        h, _ = self.gru_2(h)

        #
        # Third linears
        #
        # (bs, ts, l1_dim) -> (bs, ts, l3_dim)
        h = F.relu(self.linear_5(h))

        # (bs, ts, l3_dim) -> (bs, ts, l4_dim)
        h = F.relu(self.linear_6(h))
        h = self.dropout_out(h)

        #
        # Output layer
        #
        # (bs, ts, hs) -> (bs, ts, 2)
        h = self.linear_out(h)
        # h = F.log_softmax(h, dim=-1)

        outputs = OrderedDict()

        if self.config.predict_target:
            outputs[const.TARGET_TAGS] = h
        if self.config.predict_gaps:
            outputs[const.GAP_TAGS] = h
        if self.config.predict_source:
            outputs[const.SOURCE_TAGS] = h

        return outputs



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from torchtext import data


class Corpus:
    def __init__(self, fields_examples=None, dataset_fields=None):
        """Create a Corpus by specifying examples and fields.
        Arguments:
            fields_examples: A list of lists of field values per example.
            dataset_fields: A list of pairs (field name, field object).
        Both lists have the same size (number of fields).
        """
        self.fields_examples = (
            fields_examples if fields_examples is not None else []
        )
        self.dataset_fields = (
            dataset_fields if dataset_fields is not None else []
        )
        self.number_of_examples = (
            len(self.fields_examples[0]) if self.fields_examples else 0
        )

    def examples_per_field(self):
        examples = {
            field: examples
            for (field, _), examples in zip(
                self.dataset_fields, self.fields_examples
            )
        }
        return examples

    @classmethod
    def from_files(cls, fields, files):
        """Create a QualityEstimationDataset given paths and fields.

        Arguments:
            fields: A dict between field name and field object.
            files: A dict between field name and file dict (with 'name' and
                   'format' keys).
        """
        fields_examples = []
        dataset_fields = []

        # first load the data for each field
        for attrib_name, field in fields.items():
            file_dict = files[attrib_name]
            file_name = file_dict['name']
            reader = file_dict['reader']
            if not reader:
                with open(file_name, 'r', encoding='utf8') as f:
                    fields_values_for_example = [line.strip() for line in f]
            else:
                fields_values_for_example = reader(file_name)
            fields_examples.append(fields_values_for_example)
            dataset_fields.append((attrib_name, field))

        # then add each corresponding sentence from each field
        nb_lines = [len(fe) for fe in fields_examples]
        assert min(nb_lines) == max(nb_lines)  # Assert files have the same size
        return cls(fields_examples, dataset_fields)

    @staticmethod
    def read_tabular_file(file_path, sep='\t', extract_column=None):
        examples = []
        line_values = []
        with open(file_path, 'r', encoding='utf8') as f:
            num_columns = None
            for line_num, line in enumerate(f):
                line = line.rstrip()
                if line:
                    values = line.split(sep)
                    line_values.append(values)
                    if num_columns is None:
                        num_columns = len(values)
                        if extract_column is not None and (
                            extract_column < 1 or extract_column > num_columns
                        ):
                            raise IndexError(
                                'Cannot extract column {} (of {})'.format(
                                    extract_column, num_columns
                                )
                            )
                    elif len(values) != num_columns:
                        raise IndexError(
                            'Number of columns ({}) in line {} is different '
                            '({}) for file: {}'.format(
                                len(values),
                                line_num + 1,
                                num_columns,
                                file_path,
                            )
                        )
                else:
                    if extract_column is not None:
                        examples.append(
                            ' '.join(
                                [
                                    values[extract_column - 1]
                                    for values in line_values
                                ]
                            )
                        )
                    else:
                        examples.append(
                            [
                                ' '.join([values[i] for values in line_values])
                                for i in range(num_columns)
                            ]
                        )

                    line_values = []
        if line_values:  # Add trailing lines before EOF.
            if extract_column is not None:
                examples.append(
                    ' '.join(
                        [values[extract_column - 1] for values in line_values]
                    )
                )
            else:
                examples.append(
                    [
                        ' '.join([values[i] for values in line_values])
                        for i in range(num_columns)
                    ]
                )

        return examples

    @classmethod
    def from_tabular_file(cls, fields, file_fields, file_path, sep='\t'):
        """Create a QualityEstimationDataset given paths and fields.

        Arguments:
            fields: A dict between field name and field object.
            file_fields: A list of field names for each column of the file
                (by order). File fields not in fields will be ignored, but
                every field in fields should correspond to some column.
            file_path: Path to the tabular file.
        """
        fields_examples = []
        dataset_fields = []
        examples = {field_name: [] for field_name in fields.keys()}
        example_values = []
        with open(file_path, 'r', encoding='utf8') as f:
            for line in f:
                line = line.rstrip()
                if line:
                    values = line.split(sep)
                    example_values.append(values)
                else:
                    for i, field_name in enumerate(file_fields):
                        if field_name not in fields:  # TODO
                            continue
                        examples[field_name].append(
                            ' '.join([values[i] for values in example_values])
                        )
                    example_values = []
        if example_values:  # Add trailing lines before EOF.
            for i, field_name in enumerate(file_fields):
                if field_name not in fields:
                    continue
                examples[field_name].append(
                    ' '.join([values[i] for values in example_values])
                )

        for attrib_name, field in fields.items():
            fields_examples.append(examples[attrib_name])
            dataset_fields.append((attrib_name, field))

        # then add each corresponding sentence from each field
        nb_lines = [len(fe) for fe in fields_examples]
        assert min(nb_lines) == max(nb_lines)  # Assert files have the same size
        return cls(fields_examples, dataset_fields)

    def __iter__(self):
        for j in range(self.number_of_examples):
            fields_values_for_example = [
                self.fields_examples[i][j]
                for i in range(len(self.dataset_fields))
            ]
            yield data.Example.fromlist(
                fields_values_for_example, self.dataset_fields
            )

    def paste_fields(self, corpus):
        """Pastes (appends) fields from another corpus.
        Arguments:
            corpus: A corpus object. Must have the same number of examples as
                the current corpus.
        """
        assert self.number_of_examples == corpus.number_of_examples
        self.fields_examples += corpus.fields_examples
        self.dataset_fields += corpus.dataset_fields



"""A generic implementation of an abstract structured linear classifier."""

#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging

import numpy as np

from .linear_model import LinearModel
from .structured_decoder import StructuredDecoder
from .utils import nearly_eq_tol

logger = logging.getLogger(__name__)


class StructuredClassifier:
    """ An abstract structured classifier."""

    def __init__(self):
        self.model = LinearModel()
        self.decoder = StructuredDecoder()
        self.use_binary_features = False
        self.feature_indices = None

    def save(self, model_path):
        """Save the full configuration and model."""
        raise NotImplementedError

    def load(self, model_path):
        """Load the full configuration and model."""
        raise NotImplementedError

    def create_instances(self, dataset):
        """Preprocess the dataset if needed to create instances.
        Default is returning the dataset itself. Override if needed."""
        return dataset

    def label_instance(self, instance, parts, predicted_output):
        """Return a labeled instance by adding the predicted output
        information."""
        raise NotImplementedError

    def create_prediction(self, instance, parts, predicted_output):
        """Create a prediction for an instance."""
        raise NotImplementedError

    def make_parts(self, instance):
        """Compute the task-specific parts for this instance."""
        raise NotImplementedError

    def make_features(self, instance, parts):
        """Create a feature vector for each part."""
        raise NotImplementedError

    def compute_scores(self, instance, parts, features):
        """Compute a score for every part in the instance using the current
        model and the part-specific features."""
        num_parts = len(parts)
        scores = np.zeros(num_parts)
        for r in range(num_parts):
            if self.use_binary_features:
                scores[r] = self.model.compute_score_binary_features(
                    features[r]
                )
            else:
                scores[r] = self.model.compute_score(features[r])
        return scores

    def run(self, instance):
        """Run the structured classifier on a single instance."""
        parts, gold_output = self.make_parts(instance)
        features = self.make_features(instance, parts)
        scores = self.compute_scores(instance, parts, features)
        predicted_output = self.decoder.decode(instance, parts, scores)
        labeled_instance = self.label_instance(
            instance, parts, predicted_output
        )
        return labeled_instance

    def test(self, instances):
        """Run the structured classifier on dev/test data."""
        num_mistakes = 0
        num_parts_total = 0
        predictions = []
        for instance in instances:
            # TODO: use self.run(instance) instead?
            parts, gold_output = self.make_parts(instance)
            features = self.make_features(instance, parts)
            scores = self.compute_scores(instance, parts, features)
            predicted_output = self.decoder.decode(instance, parts, scores)
            predictions.append(
                self.create_prediction(instance, parts, predicted_output)
            )
            num_parts = len(parts)
            assert len(predicted_output) == num_parts
            assert len(gold_output) == num_parts
            for i in range(num_parts):
                if not nearly_eq_tol(gold_output[i], predicted_output[i], 1e-6):
                    num_mistakes += 1
            num_parts_total += num_parts

        logger.info(
            'Part accuracy: %f',
            float(num_parts_total - num_mistakes) / float(num_parts_total),
        )
        return predictions

    def evaluate(self, instances, predictions, print_scores=True):
        """Evaluate the structure classifier, computing a task-dependent
        evaluation metric."""
        raise NotImplementedError



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from pathlib import Path
from pprint import pformat

import torch

from kiwi import constants as const
from kiwi.cli.pipelines.train import build_parser
from kiwi.data import builders, utils
from kiwi.data.iterators import build_bucket_iterator
from kiwi.data.utils import (
    save_training_datasets,
    save_vocabularies_from_datasets,
)
from kiwi.lib.utils import (
    configure_logging,
    configure_seed,
    merge_namespaces,
    save_args_to_file,
    setup_output_directory,
)
from kiwi.loggers import tracking_logger
from kiwi.models.linear_word_qe_classifier import LinearWordQEClassifier
from kiwi.models.model import Model
from kiwi.trainers.callbacks import Checkpoint
from kiwi.trainers.linear_word_qe_trainer import LinearWordQETrainer
from kiwi.trainers.trainer import Trainer
from kiwi.trainers.utils import optimizer_class

logger = logging.getLogger(__name__)


class TrainRunInfo:
    """
    Encapsulates relevant information on training runs.

    Can be instantiated with a trainer object.

    Attributes:
        stats: Stats of the best model so far
        model_path: Path of the best model so far
        run_uuid: Unique identifier of the current run
    """

    def __init__(self, trainer):
        # FIXME: linear trainer not yet supported here
        #   (no full support to checkpointer)
        self.stats = trainer.checkpointer.best_stats()
        self.model_path = trainer.checkpointer.best_model_path()
        self.run_uuid = tracking_logger.run_uuid


def train_from_file(filename):
    """
    Loads options from a config file and calls the training procedure.

    Args:
        filename (str): filename of the configuration file
    """
    parser = build_parser()
    options = parser.parse_config_file(filename)
    return train_from_options(options)


def train_from_options(options):
    """
    Runs the entire training pipeline using the configuration options received.

    These options include the pipeline and model options plus the model's API.

    Args:
        options (Namespace): All the configuration options retrieved
            from either a config file or input flags and the model
            being used.
    """
    if options is None:
        return

    pipeline_options = options.pipeline
    model_options = options.model
    ModelClass = options.model_api

    tracking_run = tracking_logger.configure(
        run_uuid=pipeline_options.run_uuid,
        experiment_name=pipeline_options.experiment_name,
        run_name=pipeline_options.run_name,
        tracking_uri=pipeline_options.mlflow_tracking_uri,
        always_log_artifacts=pipeline_options.mlflow_always_log_artifacts,
    )

    with tracking_run:
        output_dir = setup(
            output_dir=pipeline_options.output_dir,
            seed=pipeline_options.seed,
            gpu_id=pipeline_options.gpu_id,
            debug=pipeline_options.debug,
            quiet=pipeline_options.quiet,
        )

        all_options = merge_namespaces(pipeline_options, model_options)
        log(
            output_dir,
            config_options=vars(all_options),
            save_config=pipeline_options.save_config,
        )

        trainer = run(ModelClass, output_dir, pipeline_options, model_options)
        train_info = TrainRunInfo(trainer)

    teardown(pipeline_options)

    return train_info


def run(ModelClass, output_dir, pipeline_options, model_options):
    """
    Implements the main logic of the training module.

    Instantiates the dataset, model class and sets their attributes according
    to the pipeline options received. Loads or creates a trainer and runs it.

    Args:
        ModelClass (Model): Python Type of the Model to train
        output_dir: Directory to save models
        pipeline_options (Namespace): Generic Train Options
            load_model: load pre-trained predictor model
            resume: load trainer state and resume training
            gpu_id: Set to non-negative integer to train on GPU
            train_batch_size: Batch Size for training
            valid_batch_size: Batch size for validation

        model_options(Namespace): Model Specific options

    Returns:
        The trainer object
    """
    model_name = getattr(ModelClass, "title", ModelClass.__name__)
    logger.info("Training the {} model".format(model_name))
    # FIXME: make sure all places use output_dir
    # del pipeline_options.output_dir
    pipeline_options.output_dir = None

    # Data step
    fieldset = ModelClass.fieldset(
        wmt18_format=model_options.__dict__.get("wmt18_format")
    )

    datasets = retrieve_datasets(
        fieldset, pipeline_options, model_options, output_dir
    )
    save_vocabularies_from_datasets(output_dir, *datasets)
    if pipeline_options.save_data:
        save_training_datasets(pipeline_options.save_data, *datasets)

    # Trainer step
    device_id = None
    if pipeline_options.gpu_id is not None and pipeline_options.gpu_id >= 0:
        device_id = pipeline_options.gpu_id

    vocabs = utils.fields_to_vocabs(datasets[0].fields)

    trainer = retrieve_trainer(
        ModelClass,
        pipeline_options,
        model_options,
        vocabs,
        output_dir,
        device_id,
    )

    logger.info(str(trainer.model))
    logger.info("{} parameters".format(trainer.model.num_parameters()))
    tracking_logger.log_param(
        "model_parameters", trainer.model.num_parameters()
    )

    # Dataset iterators
    train_iter = build_bucket_iterator(
        datasets[0],
        batch_size=pipeline_options.train_batch_size,
        is_train=True,
        device=device_id,
    )
    valid_iter = build_bucket_iterator(
        datasets[1],
        batch_size=pipeline_options.valid_batch_size,
        is_train=False,
        device=device_id,
    )

    trainer.run(train_iter, valid_iter, epochs=pipeline_options.epochs)

    return trainer


def retrieve_trainer(
    ModelClass, pipeline_options, model_options, vocabs, output_dir, device_id
):
    """
    Creates a Trainer object with an associated model.

    This object encapsulates the logic behind training the model and
    checkpointing. This method uses the received pipeline options to
    instantiate a Trainer object with the the requested model and
    hyperparameters.

    Args:
        ModelClass
        pipeline_options (Namespace): Generic training options
            resume (bool): Set to true if resuming an existing run.
            load_model (str): Directory containing model.torch for loading
                pre-created model.
            checkpoint_save (bool): Boolean indicating if snapshots should be
                saved after validation runs. warning: if false, will never save
                the model.
            checkpoint_keep_only_best (int): Indicates kiwi to keep the best
                `n` models.
            checkpoint_early_stop_patience (int): Stops training if metrics
                don't improve after `n` validation runs.
            checkpoint_validation_steps (int): Perform validation every `n`
                training steps.
            optimizer (string): The optimizer to be used in training.
            learning_rate (float): Starting learning rate.
            learning_rate_decay (float): Factor of learning rate decay.
            learning_rate_decay_start (int): Start decay after epoch `x`.
            log_interval (int): Log after `k` batches.
        model_options (Namespace): Model specific options.
        vocabs (dict): Vocab dictionary.
        output_dir (str or Path): Output directory for models and stats
            concerning training.
        device_id (int): The gpu id to be used in training. Set to negative
            to use cpu.
    Returns:
        Trainer

    """

    if pipeline_options.resume:
        return Trainer.resume(local_path=output_dir, device_id=device_id)

    if pipeline_options.load_model:
        model = Model.create_from_file(pipeline_options.load_model)
    else:
        model = ModelClass.from_options(vocabs=vocabs, opts=model_options)

    checkpointer = Checkpoint(
        output_dir,
        pipeline_options.checkpoint_save,
        pipeline_options.checkpoint_keep_only_best,
        pipeline_options.checkpoint_early_stop_patience,
        pipeline_options.checkpoint_validation_steps,
    )

    if isinstance(model, LinearWordQEClassifier):
        trainer = LinearWordQETrainer(
            model,
            model_options.training_algorithm,
            model_options.regularization_constant,
            checkpointer,
        )
    else:
        # Set GPU or CPU; has to be before instantiating the optimizer
        model.to(device_id)

        # Optimizer
        OptimizerClass = optimizer_class(pipeline_options.optimizer)
        optimizer = OptimizerClass(
            model.parameters(), lr=pipeline_options.learning_rate
        )
        scheduler = None
        if 0.0 < pipeline_options.learning_rate_decay < 1.0:
            scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                optimizer,
                factor=pipeline_options.learning_rate_decay,
                patience=pipeline_options.learning_rate_decay_start,
                verbose=True,
                mode="max",
            )

        trainer = Trainer(
            model,
            optimizer,
            checkpointer,
            log_interval=pipeline_options.log_interval,
            scheduler=scheduler,
        )
    return trainer


def retrieve_datasets(fieldset, pipeline_options, model_options, output_dir):
    """
    Creates `Dataset` objects for the training and validation sets.

    Parses files according to pipeline and model options.

    Args:
        fieldset
        pipeline_options (Namespace): Generic training options
            load_data (str): Input directory for loading preprocessed data
                files.
            load_model (str): Directory containing model.torch for loading
                pre-created model.
            resume (boolean): Indicates if you should resume training from a
                previous run.
            load_vocab (str): Directory containing vocab.torch file to be
                loaded.
        model_options (Namespace): Model specific options.
        output_dir (str): Path to directory where experiment files should be
            saved.

    Returns:
        datasets (Dataset): Training and validation datasets
    """
    if pipeline_options.load_data:
        datasets = utils.load_training_datasets(
            pipeline_options.load_data, fieldset
        )
    else:
        load_vocab = None

        if pipeline_options.resume:
            load_vocab = Path(output_dir, const.VOCAB_FILE)
        elif pipeline_options.load_model:
            load_vocab = pipeline_options.load_model
        elif model_options.__dict__.get("load_pred_source"):
            load_vocab = model_options.load_pred_source
        elif model_options.__dict__.get("load_pred_target"):
            load_vocab = model_options.load_pred_target
        elif pipeline_options.load_vocab:
            load_vocab = pipeline_options.load_vocab

        datasets = builders.build_training_datasets(
            fieldset, load_vocab=load_vocab, **vars(model_options)
        )
    return datasets


def setup(output_dir, seed=42, gpu_id=None, debug=False, quiet=False):
    """
    Analyzes pipeline options and sets up requirements for running the training
    pipeline.

    This includes setting up the output directory, random seeds and the
    device(s) where training is run.

    Args:
        output_dir: Path to directory to use or None, in which case one is
            created automatically.
        seed (int): Random seed for all random engines (Python, PyTorch, NumPy).
        gpu_id (int): GPU number to use or `None` to use the CPU.
        debug (bool): Whether to increase the verbosity of output messages.
        quiet (bool): Whether to decrease the verbosity of output messages.
            Takes precedence over `debug`.

    Returns:
        output_dir(str): Path to output directory
    """
    output_dir = setup_output_directory(
        output_dir,
        tracking_logger.run_uuid,
        tracking_logger.experiment_id,
        create=True,
    )
    configure_logging(output_dir=output_dir, debug=debug, quiet=quiet)
    configure_seed(seed)

    logging.info("This is run ID: {}".format(tracking_logger.run_uuid))
    logging.info(
        "Inside experiment ID: {} ({})".format(
            tracking_logger.experiment_id, tracking_logger.experiment_name
        )
    )
    logging.info("Local output directory is: {}".format(output_dir))
    logging.info(
        "Logging execution to MLflow at: {}".format(
            tracking_logger.get_tracking_uri()
        )
    )

    if gpu_id is not None and gpu_id >= 0:
        torch.cuda.set_device(gpu_id)
        logging.info("Using GPU: {}".format(gpu_id))
    else:
        logging.info("Using CPU")

    logging.info(
        "Artifacts location: {}".format(tracking_logger.get_artifact_uri())
    )

    return output_dir


def teardown(options):
    """
    Tears down after executing prediction pipeline.

    Args:
        options(Namespace): Pipeline specific options
    """
    pass


def log(
    output_dir,
    config_options,
    config_file_name="train_config.yml",
    save_config=None,
):
    """
    Logs configuration options for the current training run.

    Args:
        output_dir (str): Path to directory where experiment files should be
            saved.
        config_options (Namespace): Namespace representing all configuration
            options.
        config_file_name (str): Filename of the config file
        save_config (str or Path): Boolean stating if you should save a
            configuration file.

    """
    logging.debug(pformat(config_options))
    config_file_copy = Path(output_dir, config_file_name)
    save_args_to_file(config_file_copy, **config_options)
    if tracking_logger.should_log_artifacts():
        tracking_logger.log_artifact(str(config_file_copy))

    if save_config:
        save_args_to_file(save_config, output_dir=output_dir, **config_options)

    # Log parameters
    tracking_logger.log_param("output_dir", output_dir)
    tracking_logger.log_param("save_config", save_config)
    for param, value in config_options.items():
        tracking_logger.log_param(param, value)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from collections import OrderedDict

import torch
import torch.nn as nn
import torch.nn.functional as F

from kiwi import constants as const
from kiwi.data.fieldsets.quetch import build_fieldset
from kiwi.metrics import CorrectMetric, F1Metric, LogMetric
from kiwi.models.model import Model, ModelConfig
from kiwi.models.utils import align_tensor, convolve_tensor, make_loss_weights


class QUETCHConfig(ModelConfig):
    def __init__(
        self,
        vocabs,
        predict_target=True,
        predict_gaps=False,
        predict_source=False,
        source_embeddings_size=50,
        target_embeddings_size=50,
        hidden_sizes=None,
        bad_weight=3.0,
        window_size=10,
        max_aligned=5,
        dropout=0.4,
        embeddings_dropout=0.4,
        freeze_embeddings=False,
    ):
        super().__init__(vocabs)

        if hidden_sizes is None:
            hidden_sizes = [100]

        source_vectors = vocabs[const.SOURCE].vectors
        target_vectors = vocabs[const.TARGET].vectors
        if source_vectors is not None:
            source_embeddings_size = source_vectors.size(1)
        if target_vectors is not None:
            target_embeddings_size = target_vectors.size(1)

        self.source_embeddings_size = source_embeddings_size
        self.target_embeddings_size = target_embeddings_size

        self.bad_weight = bad_weight
        self.dropout = dropout
        self.embeddings_dropout = embeddings_dropout
        self.freeze_embeddings = freeze_embeddings
        # self.predict_side = predict_side

        # if predicting tags or source, default predict_target=true
        # doesn't make sense
        if predict_gaps or predict_source:
            predict_target = predict_target

        self.predict_target = predict_target
        self.predict_gaps = predict_gaps
        self.predict_source = predict_source

        self.window_size = window_size
        self.max_aligned = max_aligned
        self.hidden_sizes = hidden_sizes

        if const.SOURCE_TAGS in vocabs:
            self.tags_pad_id = vocabs[const.SOURCE_TAGS].stoi[const.PAD]
        elif const.GAP_TAGS in vocabs:
            self.tags_pad_id = vocabs[const.GAP_TAGS].stoi[const.PAD]
        else:
            self.tags_pad_id = vocabs[const.TARGET_TAGS].stoi[const.PAD]
        # FIXME: this might not correspond to reality (in vocabs)!
        self.nb_classes = len(const.LABELS)
        self.tag_bad_index = const.BAD_ID
        self.pad_token = const.PAD
        self.unaligned_idx = const.UNALIGNED_ID
        self.source_padding_idx = const.PAD_ID
        self.target_padding_idx = const.PAD_ID


@Model.register_subclass
class QUETCH(Model):
    """QUality Estimation from scraTCH (QUETCH) model.

    TODO: add references.

    """

    title = "QUETCH"

    def __init__(self, vocabs, **kwargs):

        super().__init__(vocabs=vocabs, ConfigCls=QUETCHConfig, **kwargs)

        self.source_emb = None
        self.target_emb = None
        self.embeddings_dropout = None
        self.linear = None
        self.dropout = None
        self.linear_out = None

        source_vectors = vocabs[const.SOURCE].vectors
        target_vectors = vocabs[const.TARGET].vectors
        self.build(source_vectors, target_vectors)

    @staticmethod
    def fieldset(*args, **kwargs):
        return build_fieldset(*args, **kwargs)

    @staticmethod
    def from_options(vocabs, opts):
        model = QUETCH(
            vocabs=vocabs,
            predict_target=opts.predict_target,
            predict_gaps=opts.predict_gaps,
            predict_source=opts.predict_source,
            source_embeddings_size=opts.source_embeddings_size,
            target_embeddings_size=opts.target_embeddings_size,
            hidden_sizes=opts.hidden_sizes,
            bad_weight=opts.bad_weight,
            window_size=opts.window_size,
            max_aligned=opts.max_aligned,
            dropout=opts.dropout,
            embeddings_dropout=opts.embeddings_dropout,
            freeze_embeddings=opts.freeze_embeddings,
        )
        return model

    def loss(self, model_out, target):
        if self.config.predict_source:
            output_name = const.SOURCE_TAGS
        elif self.config.predict_gaps:
            output_name = const.GAP_TAGS
        else:
            output_name = const.TARGET_TAGS

        # (bs*ts, nb_classes)
        probs = model_out[output_name]
        # (bs*ts, )
        y = getattr(target, output_name)

        predicted = probs.view(-1, self.config.nb_classes)
        y = y.view(-1)

        loss = self._loss(predicted, y)
        return {const.LOSS: loss}

    def _build_embeddings(self, source_vectors=None, target_vectors=None):
        # Embeddings layers:
        if source_vectors is not None:
            # source_embeddings_size = self.source_embeddings.size(1)
            self.source_emb = nn.Embedding(
                num_embeddings=source_vectors.size(0),
                embedding_dim=source_vectors.size(1),
                padding_idx=self.config.source_padding_idx,
                _weight=source_vectors,
            )
        else:
            self.source_emb = nn.Embedding(
                num_embeddings=self.config.source_vocab_size,
                embedding_dim=self.config.source_embeddings_size,
                padding_idx=self.config.source_padding_idx,
            )
        if target_vectors is not None:
            self.target_emb = nn.Embedding(
                num_embeddings=target_vectors.size(0),
                embedding_dim=target_vectors.size(1),
                padding_idx=self.config.target_padding_idx,
                _weight=target_vectors,
            )
        else:
            self.target_emb = nn.Embedding(
                num_embeddings=self.config.target_vocab_size,
                embedding_dim=self.config.target_embeddings_size,
                padding_idx=self.config.target_padding_idx,
            )
        if self.config.freeze_embeddings:
            self.source_emb.weight.requires_grad = False
            self.source_emb.bias.requires_grad = False
            self.target_emb.weight.requires_grad = False
            self.target_emb.bias.requires_grad = False

        self.embeddings_dropout = nn.Dropout(self.config.embeddings_dropout)

    def build(self, source_vectors=None, target_vectors=None):

        hidden_size = self.config.hidden_sizes[0]
        nb_classes = self.config.nb_classes
        dropout = self.config.dropout

        weight = make_loss_weights(
            nb_classes, const.BAD_ID, self.config.bad_weight
        )

        self._loss = nn.CrossEntropyLoss(
            weight=weight, ignore_index=const.PAD_TAGS_ID
        )

        # Embeddings layers:
        self._build_embeddings(source_vectors, target_vectors)

        feature_set_size = (
            self.config.source_embeddings_size
            + self.config.target_embeddings_size
        ) * self.config.window_size

        self.linear = nn.Linear(feature_set_size, hidden_size)
        self.linear_out = nn.Linear(hidden_size, nb_classes)

        self.dropout = nn.Dropout(dropout)

        torch.nn.init.xavier_uniform_(self.linear.weight)
        torch.nn.init.xavier_uniform_(self.linear_out.weight)
        torch.nn.init.constant_(self.linear.bias, 0.0)
        torch.nn.init.constant_(self.linear_out.bias, 0.0)

        self.is_built = True

    def make_input(self, batch, side):
        target_input, target_lengths = getattr(batch, const.TARGET)
        source_input, source_lengths = getattr(batch, const.SOURCE)
        alignments = batch.alignments

        if self.config.predict_gaps and not self.config.predict_target:
            target_input = F.pad(
                target_input,
                pad=(0, 1),
                value=self.vocabs[const.TARGET].stoi[const.UNALIGNED],
            )
            source_input = F.pad(
                source_input,
                pad=(0, 1),
                value=self.vocabs[const.SOURCE].stoi[const.UNALIGNED],
            )

        target_input = convolve_tensor(
            target_input,
            self.config.window_size,
            self.config.target_padding_idx,
        )

        source_input = convolve_tensor(
            source_input,
            self.config.window_size,
            self.config.source_padding_idx,
        )

        if side == const.SOURCE_TAGS:
            alignments = [
                [alignment[::-1] for alignment in example_alignment]
                for example_alignment in alignments
            ]
            target_input, nb_alignments = align_tensor(
                target_input,
                alignments,
                self.config.max_aligned,
                self.config.unaligned_idx,
                self.config.target_padding_idx,
                pad_size=source_input.shape[1],
            )
        else:
            source_input, nb_alignments = align_tensor(
                source_input,
                alignments,
                self.config.max_aligned,
                self.config.unaligned_idx,
                self.config.source_padding_idx,
                pad_size=target_input.shape[1],
            )

        return target_input, source_input, nb_alignments

    def forward(self, batch):
        assert self.is_built

        if self.config.predict_source:
            align_side = const.SOURCE_TAGS
        else:
            align_side = const.TARGET_TAGS

        target_input, source_input, nb_alignments = self.make_input(
            batch, align_side
        )

        #
        # Source Branch
        #
        # (bs, ts, aligned, window) -> (bs, ts, aligned, window, emb)
        h_source = self.source_emb(source_input)

        if len(h_source.shape) == 5:
            # (bs, ts, aligned, window, emb) -> (bs, ts, window, emb)
            h_source = h_source.sum(2, keepdim=False) / nb_alignments.unsqueeze(
                -1
            ).unsqueeze(-1)

        # (bs, ts, window, emb) -> (bs, ts, window * emb)
        h_source = h_source.view(source_input.size(0), source_input.size(1), -1)

        #
        # Target Branch
        #
        # (bs, ts * window) -> (bs, ts * window, emb)
        h_target = self.target_emb(target_input)

        if len(h_target.shape) == 5:
            # (bs, ts, aligned, window, emb) -> (bs, ts, window, emb)
            h_target = h_target.sum(2, keepdim=False) / nb_alignments.unsqueeze(
                -1
            ).unsqueeze(-1)

        # (bs, ts * window, emb) -> (bs, ts, window * emb)
        h_target = h_target.view(target_input.size(0), target_input.size(1), -1)

        #
        # POS tags branches
        #
        feature_set = (h_source, h_target)

        #
        # Merge Branches
        #
        # (bs, ts, window * emb) -> (bs, ts, 2 * window * emb)
        h = torch.cat(feature_set, dim=-1)
        h = self.embeddings_dropout(h)
        # (bs, ts, 2 * window * emb) -> (bs, ts, hs)
        h = torch.tanh(self.linear(h))
        h = self.dropout(h)

        # (bs, ts, hs) -> (bs, ts, 2)
        h = self.linear_out(h)

        outputs = OrderedDict()

        if self.config.predict_target:
            outputs[const.TARGET_TAGS] = h
        if self.config.predict_gaps:
            outputs[const.GAP_TAGS] = h
        if self.config.predict_source:
            outputs[const.SOURCE_TAGS] = h

        return outputs

    @staticmethod
    def _unmask(tensor, mask):
        lengths = mask.int().sum(dim=-1)
        return [x[: lengths[i]] for i, x in enumerate(tensor)]

    def metrics(self):
        metrics = []

        if self.config.predict_target:
            metrics.append(
                F1Metric(
                    prefix=const.TARGET_TAGS,
                    target_name=const.TARGET_TAGS,
                    PAD=const.PAD_TAGS_ID,
                    labels=const.LABELS,
                )
            )
            metrics.append(
                CorrectMetric(
                    prefix=const.TARGET_TAGS,
                    target_name=const.TARGET_TAGS,
                    PAD=const.PAD_TAGS_ID,
                )
            )
        if self.config.predict_source:
            metrics.append(
                F1Metric(
                    prefix=const.SOURCE_TAGS,
                    target_name=const.SOURCE_TAGS,
                    PAD=const.PAD_TAGS_ID,
                    labels=const.LABELS,
                )
            )
            metrics.append(
                CorrectMetric(
                    prefix=const.SOURCE_TAGS,
                    target_name=const.SOURCE_TAGS,
                    PAD=const.PAD_TAGS_ID,
                )
            )
        if self.config.predict_gaps:
            metrics.append(
                F1Metric(
                    prefix=const.GAP_TAGS,
                    target_name=const.GAP_TAGS,
                    PAD=const.PAD_TAGS_ID,
                    labels=const.LABELS,
                )
            )
            metrics.append(
                CorrectMetric(
                    prefix=const.GAP_TAGS,
                    target_name=const.GAP_TAGS,
                    PAD=const.PAD_TAGS_ID,
                )
            )

        metrics.append(LogMetric(targets=[(const.LOSS, const.LOSS)]))

        return metrics

    def metrics_ordering(self):
        return max



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from functools import partial

import torch
from torchtext.vocab import Vectors

from kiwi.constants import PAD, START, STOP, UNK

logger = logging.getLogger(__name__)


class WordEmbeddings(Vectors):
    def __init__(
        self,
        name,
        emb_format='polyglot',
        binary=True,
        map_fn=lambda x: x,
        **kwargs
    ):
        """
        Arguments:
           emb_format: the saved embedding model format, choices are:
                       polyglot, word2vec, fasttext, glove and text
           binary: only for word2vec, fasttext and text
           map_fn: a function that maps special original tokens
                       to Polyglot tokens (e.g. <eos> to </S>)
           save_vectors: save a vectors cache
        """
        self.binary = binary
        self.emb_format = emb_format

        self.itos = None
        self.stoi = None
        self.dim = None
        self.vectors = None

        self.map_fn = map_fn
        super().__init__(name, **kwargs)

    def __getitem__(self, token):
        if token in self.stoi:
            token = self.map_fn(token)
            return self.vectors[self.stoi[token]]
        else:
            return self.unk_init(torch.Tensor(1, self.dim))

    def cache(self, name, cache, url=None, max_vectors=None):
        if self.emb_format in ['polyglot', 'glove']:
            try:
                from polyglot.mapping import Embedding
            except ImportError:
                logger.error('Please install `polyglot` package first.')
                return None
            if self.emb_format == 'polyglot':
                embeddings = Embedding.load(name)
            else:
                embeddings = Embedding.from_glove(name)
            self.itos = embeddings.vocabulary.id_word
            self.stoi = embeddings.vocabulary.word_id
            self.dim = embeddings.shape[1]
            self.vectors = torch.Tensor(embeddings.vectors).view(-1, self.dim)

        elif self.emb_format in ['word2vec', 'fasttext']:
            try:
                from gensim.models import KeyedVectors
            except ImportError:
                logger.error('Please install `gensim` package first.')
                return None
            embeddings = KeyedVectors.load_word2vec_format(
                name, unicode_errors='ignore', binary=self.binary
            )
            self.itos = embeddings.index2word
            self.stoi = dict(zip(self.itos, range(len(self.itos))))
            self.dim = embeddings.vector_size
            self.vectors = torch.Tensor(embeddings.vectors).view(-1, self.dim)

        elif self.emb_format == 'text':
            tokens = []
            vectors = []
            if self.binary:
                import pickle

                # vectors should be a dict mapping str keys to numpy arrays
                with open(name, 'rb') as f:
                    d = pickle.load(f)
                    tokens = list(d.keys())
                    vectors = list(d.values())
            else:
                # each line should contain a token and its following fields
                # <token> <vector_value_1> ... <vector_value_n>
                with open(name, 'r', encoding='utf8') as f:
                    for line in f:
                        if line:  # ignore empty lines
                            fields = line.rstrip().split()
                            tokens.append(fields[0])
                            vectors.append(list(map(float, fields[1:])))
            self.itos = tokens
            self.stoi = dict(zip(self.itos, range(len(self.itos))))
            self.vectors = torch.Tensor(vectors)
            self.dim = self.vectors.shape[1]


def map_to_polyglot(token):
    mapping = {UNK: '<UNK>', PAD: '<PAD>', START: '<S>', STOP: '</S>'}
    if token in mapping:
        return mapping[token]
    return token


Polyglot = partial(
    WordEmbeddings, emb_format='polyglot', map_fn=map_to_polyglot
)
Word2Vec = partial(WordEmbeddings, emb_format='word2vec')
FastText = partial(WordEmbeddings, emb_format='fasttext')
Glove = partial(WordEmbeddings, emb_format='glove')
TextVectors = partial(WordEmbeddings, emb_format='text')

AvailableVectors = {
    'polyglot': Polyglot,
    'word2vec': Word2Vec,
    'fasttext': FastText,
    'glove': Glove,
    'text': TextVectors,
}



"""Decoder for word-level quality estimation."""

#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import numpy as np

from .sequence_parts import SequenceBigramPart, SequenceUnigramPart
from .structured_decoder import StructuredDecoder


def logzero():
    """Return log of zero."""
    return -np.inf


class LinearWordQEDecoder(StructuredDecoder):
    """A decoder for word-level quality estimation."""

    def __init__(
        self, estimator, cost_false_positives=0.5, cost_false_negatives=0.5
    ):
        StructuredDecoder.__init__(self)
        self.estimator = estimator
        self.cost_false_positives = cost_false_positives
        self.cost_false_negatives = cost_false_negatives

    def decode_mira(
        self, instance, parts, scores, gold_outputs, old_mira=False
    ):
        """Cost-augmented decoder. Allows a compromise between precision and
        recall. In general:
        p = a - (a+b)*z0
        q = b*sum(z0)
        p'*z + q = a*sum(z) - (a+b)*z0'*z + b*sum(z0)
                 = a*(1-z0)'*z + b*(1-z)'*z0
        a => penalty for predicting 1 when it is 0 (FP)
        b => penalty for predicting 0 when it is 1 (FN)

        F1: a = 0.5, b = 0.5
        recall: a = 0, b = 1"""

        a = self.cost_false_positives
        b = self.cost_false_negatives

        # Allow multiple bad labels.
        bad = []
        for label in self.estimator.labels:
            coarse_label = self.estimator.get_coarse_label(label)
            if coarse_label == 'BAD':
                bad.append(self.estimator.labels[label])
        bad = set(bad)
        index_parts = [
            i
            for i in range(len(parts))
            if isinstance(parts[i], SequenceUnigramPart)
            and parts[i].label in bad
        ]
        p = np.zeros(len(parts))
        p[index_parts] = a - (a + b) * gold_outputs[index_parts]
        q = b * np.ones(len(gold_outputs[index_parts])).dot(
            gold_outputs[index_parts]
        )
        if old_mira:
            predicted_outputs = self.decode(instance, parts, scores)
        else:
            scores_cost = scores + p
            predicted_outputs = self.decode(instance, parts, scores_cost)
        cost = p.dot(predicted_outputs) + q
        loss = cost + scores.dot(predicted_outputs - gold_outputs)
        return predicted_outputs, cost, loss

    def decode(self, instance, parts, scores):
        """Decoder. Return the most likely sequence of OK/BAD labels."""
        if self.estimator.use_bigrams:
            return self.decode_with_bigrams(instance, parts, scores)
        else:
            return self.decode_with_unigrams(instance, parts, scores)

    def decode_with_unigrams(self, instance, parts, scores):
        """Decoder for a non-sequential model (unigrams only)."""
        predicted_output = np.zeros(len(scores))
        parts_by_index = [[] for _ in range(instance.num_words())]
        for r, part in enumerate(parts):
            parts_by_index[part.index].append(r)

        for i in range(instance.num_words()):
            num_labels = len(parts_by_index[i])
            label_scores = np.zeros(num_labels)
            predicted_for_word = [0] * num_labels
            for k, r in enumerate(parts_by_index[i]):
                label_scores[k] = scores[r]
            best = np.argmax(label_scores)
            predicted_for_word[best] = 1.0
            r = parts_by_index[i][best]
            predicted_output[r] = 1.0

        return predicted_output

    def decode_with_bigrams(self, instance, parts, scores):
        """Decoder for a sequential model (with bigrams)."""
        num_labels = len(self.estimator.labels)
        num_words = instance.num_words()
        initial_scores = np.zeros(num_labels)
        transition_scores = np.zeros((num_words - 1, num_labels, num_labels))
        final_scores = np.zeros(num_labels)
        emission_scores = np.zeros((num_words, num_labels))

        indexed_unigram_parts = [{} for _ in range(num_words)]
        indexed_bigram_parts = [{} for _ in range(num_words + 1)]

        for r, part in enumerate(parts):
            if isinstance(part, SequenceUnigramPart):
                indexed_unigram_parts[part.index][part.label] = r
                emission_scores[part.index, part.label] = scores[r]
            elif isinstance(part, SequenceBigramPart):
                indexed_bigram_parts[part.index][
                    (part.label, part.previous_label)
                ] = r
                if part.previous_label < 0:
                    assert part.index == 0
                    initial_scores[part.label] = scores[r]
                elif part.label < 0:
                    assert part.index == num_words
                    final_scores[part.previous_label] = scores[r]
                else:
                    transition_scores[
                        part.index - 1, part.label, part.previous_label
                    ] = scores[r]
            else:
                raise NotImplementedError

        best_path, _ = self.run_viterbi(
            initial_scores, transition_scores, final_scores, emission_scores
        )

        predicted_output = np.zeros(len(scores))
        previous_label = -1
        for i, label in enumerate(best_path):
            r = indexed_unigram_parts[i][label]
            predicted_output[r] = 1.0
            r = indexed_bigram_parts[i][(label, previous_label)]
            predicted_output[r] = 1.0
            previous_label = label
        r = indexed_bigram_parts[num_words][(-1, previous_label)]
        predicted_output[r] = 1.0

        return predicted_output

    def run_viterbi(
        self, initial_scores, transition_scores, final_scores, emission_scores
    ):
        """Computes the viterbi trellis for a given sequence.
        Receives:
        - Initial scores: (num_states) array
        - Transition scores: (length-1, num_states, num_states) array
        - Final scores: (num_states) array
        - Emission scores: (length, num_states) array."""

        length = np.size(emission_scores, 0)  # Length of the sequence.
        num_states = np.size(initial_scores)  # Number of states.

        # Variables storing the Viterbi scores.
        viterbi_scores = np.zeros([length, num_states]) + logzero()

        # Variables storing the paths to backtrack.
        viterbi_paths = -np.ones([length, num_states], dtype=int)

        # Most likely sequence.
        best_path = -np.ones(length, dtype=int)

        # Initialization.
        viterbi_scores[0, :] = emission_scores[0, :] + initial_scores

        # Viterbi loop.
        for pos in range(1, length):
            for current_state in range(num_states):
                viterbi_scores[pos, current_state] = np.max(
                    viterbi_scores[pos - 1, :]
                    + transition_scores[pos - 1, current_state, :]
                )
                viterbi_scores[pos, current_state] += emission_scores[
                    pos, current_state
                ]
                viterbi_paths[pos, current_state] = np.argmax(
                    viterbi_scores[pos - 1, :]
                    + transition_scores[pos - 1, current_state, :]
                )
        # Termination.
        assert len(viterbi_scores[length - 1, :] + final_scores)
        best_score = np.max(viterbi_scores[length - 1, :] + final_scores)
        best_path[length - 1] = np.argmax(
            viterbi_scores[length - 1, :] + final_scores
        )

        # Backtrack.
        for pos in range(length - 2, -1, -1):
            best_path[pos] = viterbi_paths[pos + 1, best_path[pos + 1]]

        return best_path, best_score



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from torchtext import data


class QEDataset(data.Dataset):
    """Defines a dataset for quality estimation. Based on the WMT 201X."""

    @staticmethod
    def sort_key(ex):
        # don't work for pack_padded_sequences
        # return data.interleave_keys(len(ex.source), len(ex.target))
        return len(ex.source)

    def __init__(self, examples, fields, filter_pred=None):
        """Create a dataset from a list of Examples and Fields.

        Arguments:
            examples: List of Examples.
            fields (List(tuple(str, Field))): The Fields to use in this tuple.
                The string is a field name, and the Field is the associated
                field.
            filter_pred (callable or None): Use only examples for which
                filter_pred(example) is True, or use all examples if None.
                Default is None.
        """
        # ensure that examples is not a generator
        examples = list(examples)
        super().__init__(examples, fields, filter_pred)

    def __getstate__(self):
        """For pickling. Copied from OpenNMT-py DatasetBase implementation.
        """
        return self.__dict__

    def __setstate__(self, _d):
        """For pickling. Copied from OpenNMT-py DatasetBase implementation.
        """
        self.__dict__.update(_d)

    def __reduce_ex__(self, proto):
        """For pickling. Copied from OpenNMT-py DatasetBase implementation.
        """
        return super(QEDataset, self).__reduce_ex__(proto)

    def split(
        self,
        split_ratio=0.7,
        stratified=False,
        strata_field='label',
        random_state=None,
    ):
        datasets = super().split(
            split_ratio, stratified, strata_field, random_state
        )
        casted_datasets = [
            QEDataset(examples=dataset.examples, fields=dataset.fields)
            for dataset in datasets
        ]
        return casted_datasets



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from kiwi.metrics import metrics

F1Metric = metrics.F1Metric
LogMetric = metrics.LogMetric
ExpectedErrorMetric = metrics.ExpectedErrorMetric
PerplexityMetric = metrics.PerplexityMetric
CorrectMetric = metrics.CorrectMetric
RMSEMetric = metrics.RMSEMetric
PearsonMetric = metrics.PearsonMetric
SpearmanMetric = metrics.SpearmanMetric
TokenMetric = metrics.TokenMetric
ThresholdCalibrationMetric = metrics.ThresholdCalibrationMetric



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from torch import optim


def optimizer_class(name):
    if name == 'sgd':
        OptimizerClass = optim.SGD
    elif name == 'adagrad':
        OptimizerClass = optim.Adagrad
    elif name == 'adadelta':
        OptimizerClass = optim.Adadelta
    elif name == 'adam':
        OptimizerClass = optim.Adam
    elif name == 'sparseadam':
        OptimizerClass = optim.SparseAdam
    else:
        raise RuntimeError("Invalid optim method: " + name)
    return OptimizerClass



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
from distutils.util import strtobool

from kiwi.cli.better_argparse import PipelineParser
from kiwi.cli.models import linear, nuqe, predictor, predictor_estimator, quetch
from kiwi.lib import train

logger = logging.getLogger(__name__)


def train_opts(parser):
    # Training loop options
    group = parser.add_argument_group('training')

    group.add_argument(
        '--epochs', type=int, default=50, help='Number of epochs for training.'
    )
    group.add_argument(
        '--train-batch-size',
        type=int,
        default=64,
        help='Maximum batch size for training.',
    )
    group.add_argument(
        '--valid-batch-size',
        type=int,
        default=64,
        help='Maximum batch size for evaluating.',
    )

    # Optimization options
    group = parser.add_argument_group('training-optimization')
    group.add_argument(
        '--optimizer',
        default='adam',
        choices=['sgd', 'adagrad', 'adadelta', 'adam', 'sparseadam'],
        help='Optimization method.',
    )
    group.add_argument(
        '--learning-rate',
        type=float,
        default=1.0,
        help='Starting learning rate. '
        'Recommended settings: sgd = 1, adagrad = 0.1, '
        'adadelta = 1, adam = 0.001',
    )
    group.add_argument(
        '--learning-rate-decay',
        type=float,
        default=1.0,
        help='Decay learning rate by this factor. ',
    )
    group.add_argument(
        '--learning-rate-decay-start',
        type=int,
        default=0,
        help='Start decay after this epoch.',
    )

    # Saving and resuming options
    group = parser.add_argument_group('training-save-load')
    group.add_argument(
        '--checkpoint-validation-steps',
        type=int,
        default=0,
        help='Perform validation every X training batches. Saves model'
        ' if `checkpoint-save` is true.',
    )
    group.add_argument(
        '--checkpoint-save',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=True,
        help='Save a training snapshot when validation is run. If false '
        'it will never save the model.',
    )
    group.add_argument(
        '--checkpoint-keep-only-best',
        type=int,
        default=1,
        help='Keep only n best models according to main metric (F1Mult '
        'by default); 0 will keep all.',
    )
    group.add_argument(
        '--checkpoint-early-stop-patience',
        type=int,
        default=0,
        help='Stop training if evaluation metrics do not improve after X '
        'validations; 0 disables this.',
    )
    group.add_argument(
        '--resume',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Resume training a previous run. '
        'If --output-dir is not none, Kiwi will load from a checkpoint folder '
        'in that location. If --output-dir is not specified, '
        'then the --run-uuid option must be set. Files are then searched '
        'under the "runs" directory. If not found, they are '
        'downloaded from the MLflow server '
        '(check the --mlflow-tracking-uri option).',
    )


def build_parser():
    return PipelineParser(
        name='train',
        model_parsers=[
            nuqe.parser_for_pipeline('train'),
            predictor_estimator.parser_for_pipeline('train'),
            predictor.parser_for_pipeline('train'),
            quetch.parser_for_pipeline('train'),
            linear.parser_for_pipeline('train'),
        ],
        options_fn=train_opts,
    )


def main(argv=None):
    parser = build_parser()
    options = parser.parse(args=argv)
    train.train_from_options(options)


if __name__ == '__main__':  # pragma: no cover
    main()  # pragma: no cover



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from collections.__init__ import Counter, OrderedDict
from itertools import chain

from torchtext import data

from kiwi.constants import PAD, START, STOP, UNALIGNED, UNK
from kiwi.data.vocabulary import Vocabulary


class QEField(data.Field):
    def __init__(
        self,
        unaligned_token=UNALIGNED,
        unk_token=UNK,
        pad_token=PAD,
        init_token=START,
        eos_token=STOP,
        **kwargs
    ):
        kwargs.setdefault('batch_first', True)
        super().__init__(**kwargs)
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.init_token = init_token
        self.eos_token = eos_token
        self.unaligned_token = unaligned_token
        self.vocab = None
        self.vocab_cls = Vocabulary

    def build_vocab(self, *args, **kwargs):
        """Add unaligned_token to the list of special symbols."""
        counter = Counter()
        sources = []
        for arg in args:
            if isinstance(arg, data.Dataset):
                sources += [
                    getattr(arg, name)
                    for name, field in arg.fields.items()
                    if field is self
                ]
            else:
                sources.append(arg)
        for sample in sources:
            for x in sample:
                if not self.sequential:
                    x = [x]
                try:
                    counter.update(x)
                except TypeError:
                    counter.update(chain.from_iterable(x))
        specials = list(
            OrderedDict.fromkeys(
                tok
                for tok in [
                    self.unk_token,
                    self.pad_token,
                    self.init_token,
                    self.eos_token,
                    self.unaligned_token,
                ]
                if tok is not None
            )
        )
        self.vocab = self.vocab_cls(counter, specials=specials, **kwargs)



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import math
import time
from collections import OrderedDict

import numpy as np
import torch
from scipy.stats.stats import pearsonr, spearmanr
from torch import nn

from kiwi import constants as const
from kiwi.metrics.functions import fscore, precision_recall_fscore_support
from kiwi.models.utils import replace_token


class Metric:
    def __init__(
        self,
        target_name=None,
        metric_name=None,
        PAD=None,
        STOP=None,
        prefix=None,
    ):
        super().__init__()
        self.reset()
        self.prefix = prefix
        self.target_name = target_name
        self.metric_name = metric_name
        self.PAD = PAD
        self.STOP = STOP

    def update(self, **kwargs):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError

    def summarize(self, **kwargs):
        raise NotImplementedError

    def get_name(self):
        return self._prefix(self.metric_name)

    def _prefix_keys(self, summary):
        if self.prefix:
            summary = OrderedDict(
                {self._prefix(key): value for key, value in summary.items()}
            )
        return summary

    def _prefix(self, key):
        if self.prefix:
            return '{}_{}'.format(self.prefix, key)
        return key

    def token_mask(self, batch):
        target = self.get_target(batch)
        if self.PAD is not None:
            return target != self.PAD
        else:
            return torch.ones(
                target.shape, dtype=torch.uint8, device=target.device
            )

    def get_target(self, batch):
        target = getattr(batch, self.target_name)
        if self.STOP is not None:
            target = replace_token(target[:, 1:-1], self.STOP, self.PAD)
        return target

    def get_token_indices(self, batch):
        mask = self.token_mask(batch)
        return mask.view(-1).nonzero().squeeze()

    def get_predictions(self, model_out):
        predictions = model_out[self.target_name]
        return predictions

    def get_target_flat(self, batch):
        target_flat = self.get_target(batch).contiguous().view(-1)
        token_indices = self.get_token_indices(batch)
        return target_flat[token_indices]

    def get_predictions_flat(self, model_out, batch):
        predictions = self.get_predictions(model_out).contiguous()
        predictions_flat = predictions.view(-1, predictions.shape[-1]).squeeze()
        token_indices = self.get_token_indices(batch)
        return predictions_flat[token_indices]

    def get_tokens(self, batch):
        return self.token_mask(batch).sum().item()


class NLLMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='NLL', **kwargs)

    def update(self, loss, batch, **kwargs):
        self.tokens += self.get_tokens(batch)
        self.nll += loss[self.target_name].item()

    def summarize(self):
        summary = {self.metric_name: self.nll / self.tokens}
        return self._prefix_keys(summary)

    def reset(self):
        self.nll = 0.0
        self.tokens = 0


class PerplexityMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='PERP', **kwargs)

    def reset(self):
        self.tokens = 0
        self.nll = 0.0

    def update(self, loss, batch, **kwargs):
        self.tokens += self.get_tokens(batch)
        self.nll += loss[self.target_name].item()

    def summarize(self):
        summary = {self.metric_name: math.e ** (self.nll / self.tokens)}
        return self._prefix_keys(summary)


class CorrectMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='CORRECT', **kwargs)

    def update(self, model_out, batch, **kwargs):
        self.tokens += self.get_tokens(batch)
        logits = self.get_predictions_flat(model_out, batch)
        target = self.get_target_flat(batch)
        _, pred = logits.max(-1)
        correct = target == pred
        correct_count = correct.sum().item()
        self.correct += correct_count

    def summarize(self):
        summary = {self.metric_name: float(self.correct) / self.tokens}
        return self._prefix_keys(summary)

    def reset(self):
        self.correct = 0
        self.tokens = 0


class F1Metric(Metric):
    def __init__(self, labels, **kwargs):
        super().__init__(metric_name='F1_MULT', **kwargs)
        self.labels = labels

    def update(self, model_out, batch, **kwargs):
        logits = self.get_predictions_flat(model_out, batch)
        target = self.get_target_flat(batch)
        _, y_hat = logits.max(-1)
        self.Y_HAT += y_hat.tolist()
        self.Y += target.tolist()

    def summarize(self):
        summary = OrderedDict()
        _, _, f1, _ = precision_recall_fscore_support(self.Y_HAT, self.Y)
        summary[self.metric_name] = np.prod(f1)
        for i, label in enumerate(self.labels):
            summary['F1_' + label] = f1[i]
        return self._prefix_keys(summary)

    def reset(self):
        self.Y = []
        self.Y_HAT = []


class PearsonMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='PEARSON', **kwargs)

    def reset(self):
        self.predictions = []
        self.target = []

    def update(self, model_out, batch, **kwargs):
        target = self.get_target_flat(batch)
        predictions = self.get_predictions_flat(model_out, batch)
        self.predictions += predictions.tolist()
        self.target += target.tolist()

    def summarize(self):
        pearson = pearsonr(self.predictions, self.target)[0]
        summary = {self.metric_name: pearson}
        return self._prefix_keys(summary)


class SpearmanMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='SPEARMAN', **kwargs)

    def reset(self):
        self.predictions = []
        self.target = []

    def update(self, model_out, batch, **kwargs):
        target = self.get_target_flat(batch)
        predictions = self.get_predictions_flat(model_out, batch)
        self.predictions += predictions.tolist()
        self.target += target.tolist()

    def summarize(self):
        spearman = spearmanr(self.predictions, self.target)[0]
        summary = {self.metric_name: spearman}
        return self._prefix_keys(summary)


class ExpectedErrorMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='ExpErr', **kwargs)

    def update(self, model_out, batch, **kwargs):
        logits = self.get_predictions_flat(model_out, batch)
        target = self.get_target_flat(batch)
        probs = nn.functional.softmax(logits, -1)
        probs = probs.gather(-1, target.unsqueeze(-1)).squeeze()
        errors = 1.0 - probs
        self.tokens += self.get_tokens(batch)
        self.expected_error += errors.sum().item()

    def summarize(self):
        summary = {self.metric_name: self.expected_error / self.tokens}
        return self._prefix_keys(summary)

    def reset(self):
        self.expected_error = 0.0
        self.tokens = 0


class TokPerSecMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='TokPerSec', **kwargs)

    def update(self, batch, **kwargs):
        self.tokens += self.get_tokens(batch)

    def summarize(self):
        summary = {self.metric_name: self.tokens / (time.time() - self.time)}
        return self._prefix_keys(summary)

    def reset(self):
        self.tokens = 0
        self.time = time.time()


class LogMetric(Metric):
    """Logs averages of values in loss, model or batch.
    """

    def __init__(self, targets, metric_name=None, **kwargs):
        self.targets = targets
        metric_name = metric_name or self._format(*targets[0])
        super().__init__(metric_name=metric_name, **kwargs)

    def update(self, **kwargs):
        self.steps += 1
        for side, target in self.targets:
            key = self._format(side, target)
            self.log[key] += kwargs[side][target].mean().item()

    def summarize(self):
        summary = {
            key: value / float(self.steps) for key, value in self.log.items()
        }
        return self._prefix_keys(summary)

    def reset(self):
        self.log = {
            self._format(side, target): 0.0 for side, target in self.targets
        }
        self.steps = 0

    def _format(self, side, target):
        return '{}_{}'.format(side, target)


class RMSEMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='RMSE', **kwargs)

    def update(self, batch, model_out, **kwargs):
        predictions = self.get_predictions_flat(model_out, batch)
        target = self.get_target_flat(batch)
        self.squared_error += ((predictions - target) ** 2).sum().item()
        self.tokens += self.get_tokens(batch)

    def summarize(self):
        rmse = math.sqrt(self.squared_error / self.tokens)
        summary = {self.metric_name: rmse}
        return self._prefix_keys(summary)

    def reset(self):
        self.squared_error = 0.0
        self.tokens = 0


class TokenMetric(Metric):
    def __init__(self, target_token=const.UNK_ID, token_name='UNK', **kwargs):
        self.target_token = target_token
        super().__init__(metric_name='UNKS', **kwargs)

    def update(self, batch, **kwargs):
        target = self.get_target_flat(batch)
        self.targets += (target == self.target_token).sum().item()
        self.tokens += self.get_tokens(batch)

    def summarize(self):
        summary = {}
        if self.tokens:
            summary = {self.metric_name: self.targets / self.tokens}
        return self._prefix_keys(summary)

    def reset(self):
        self.tokens = 0
        self.targets = 0


class ThresholdCalibrationMetric(Metric):
    def __init__(self, **kwargs):
        super().__init__(metric_name='F1_CAL', **kwargs)

    def update(self, model_out, batch, **kwargs):
        logits = self.get_predictions_flat(model_out, batch)
        bad_probs = nn.functional.softmax(logits, -1)[:, const.BAD_ID]
        target = self.get_target_flat(batch)
        self.scores += bad_probs.tolist()
        self.Y += target.tolist()

    def summarize(self):
        summary = {}
        mid = len(self.Y) // 2
        if mid:
            perm = np.random.permutation(len(self.Y))
            self.Y = [self.Y[idx] for idx in perm]
            self.scores = [self.scores[idx] for idx in perm]
            m = MovingF1()
            fscore, threshold = m.choose(
                m.eval(self.scores[:mid], self.Y[:mid])
            )
            predictions = [
                const.BAD_ID if score >= threshold else const.OK_ID
                for score in self.scores[mid:]
            ]
            _, _, f1, _ = precision_recall_fscore_support(
                predictions, self.Y[mid:]
            )
            f1_mult = np.prod(f1)
            summary = {self.metric_name: f1_mult}
        return self._prefix_keys(summary)

    def reset(self):
        self.scores = []
        self.Y = []


class MovingMetric:
    """Class to compute the changes in one metric as a function of a second metric.
       Example: F1 score vs. Classification Threshold, Quality vs Skips
    """

    def eval(self, scores, labels):
        """Compute the graph metric1 vs metric2
        Args:
           Scores: Model Outputs
           Labels: Corresponding Labels
        """
        self.init(scores, labels)
        scores, labels = self.sort(scores, labels)
        init_threshold = scores[0]
        thresholds = [(self.compute(), init_threshold)]
        for score, label in zip(scores, labels):
            self.update(score, label)
            thresholds.append((self.compute(), score))
        return thresholds

    def init(self, scores, labels):
        """Initialize the Metric for threshold < min(scores)
        """
        return scores, labels

    def sort(self, scores, labels):
        """Sort List of labels and scores.
        """
        return zip(*sorted(zip(scores, labels)))

    def update(self, score, label):
        """Move the threshold past score
        """
        return None

    def compute(self):
        """Compute the current Value of the metric
        """
        pass

    def choose(self, thresholds):
        """Choose the best (threshold, metric) tuple from an iterable.
        """
        pass


class MovingF1(MovingMetric):
    def init(self, scores, labels, class_idx=1):
        """
        Compute F1 Mult for all decision thresholds over (scores, labels)
        Initialize the threshold s.t. all examples are classified as
        `class_idx`.
        Args:
           scores: Likelihood scores for class index
           Labels: Gold Truth classes in {0,1}
           class_index: ID of class
        """
        # -1 if class_idx == 0 , 1 if class_idx == 1
        self.sign = 2 * class_idx - 1
        class_one = sum(labels)
        class_zero = len(labels) - class_one
        self.fp_zero = (1 - class_idx) * class_one
        self.tp_zero = (1 - class_idx) * class_zero
        self.fp_one = class_idx * class_zero
        self.tp_one = class_idx * class_one

    def update(self, score, label):
        """Move the decision threshold.
        """
        self.tp_zero += self.sign * (1 - label)
        self.fp_zero += self.sign * label
        self.tp_one -= self.sign * label
        self.fp_one -= self.sign * (1 - label)

    def compute(self):
        f1_zero = fscore(self.tp_zero, self.fp_zero, self.fp_one)
        f1_one = fscore(self.tp_one, self.fp_one, self.fp_zero)
        return f1_one * f1_zero

    def choose(self, thresholds):
        return max(thresholds)


class MovingSkipsAtQuality(MovingMetric):
    """Computes Quality of skipped examples vs fraction of skips.
    """

    def __init__(
        self, scores_higher_is_better=False, labels_higher_is_better=False
    ):
        """
          Args:
          scores_higher_is_better:
              If True, higher model outputs indicate higher quality.
          labels_higher_is_better:
              If True, higher label values indicate higher quality.
        """
        self.scores_higher_is_better = scores_higher_is_better
        self.labels_higher_is_better = labels_higher_is_better

    def eval(self, scores, labels):
        """
         Args:
          scores: Model output quality or error scores. If quality scores
              are provided, pass scores_higher_is_better=True.
          labels: Ground truth quality or error scores. If quality scores
              are provided, pass labels_higher_is_better=True.
        """
        return super().eval(scores, labels)

    def init(self, scores, labels):
        """
         Args:
          scores: Model output quality or error scores. If quality scores
              are provided, pass scores_higher_is_better=True.
          labels: Ground truth quality or error scores. If quality scores
              are provided, pass labels_higher_is_better=True.
        """
        self.cumulative_qual = 0.0
        self.skipped = 0
        self.data_size = len(scores)

    def update(self, score, label):
        self.cumulative_qual += label
        self.skipped += 1

    def compute(self):
        if not self.skipped:
            return None, 0.0
        return (
            self.skipped / self.data_size,
            self.cumulative_qual / self.skipped,
        )

    def choose(self, thresholds, target_qual):
        """Chooses the smallest threshold such that
           avg. quality  is greater than or equal to target_qual
        """
        best = None
        sign = 1 if self.labels_higher_is_better else -1
        for ((skip, qual), t) in thresholds:
            if (sign * (qual - target_qual)) >= 0:
                # The quality at threshold t is admissible given target_qual
                if best is None:
                    best = ((skip, qual), t)
                else:
                    last_best = abs(best[0][0] - target_qual)
                    if abs(qual - target_qual) < last_best:
                        # The quality at threshold t is admissible given
                        # target_qual and closer than the previous best
                        best = ((skip, qual), t)
        return best

    def sort(self, scores, labels):
        return zip(
            *sorted(zip(scores, labels), reverse=self.scores_higher_is_better)
        )



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import string


class LinearWordQETokenFeatures(object):
    def __init__(
        self,
        stacked_features=None,
        source_token_count=-1,
        target_token_count=-1,
        source_target_token_count_ratio=0.0,
        token='',
        left_context='',
        right_context='',
        first_aligned_token='',
        left_alignment='',
        right_alignment='',
        is_stopword=False,
        is_punctuation=False,
        is_proper_noun=False,
        is_digit=False,
        highest_order_ngram_left=-1,
        highest_order_ngram_right=-1,
        backoff_behavior_left=0.0,
        backoff_behavior_middle=0.0,
        backoff_behavior_right=0.0,
        source_highest_order_ngram_left=-1,
        source_highest_order_ngram_right=-1,
        pseudo_reference=False,
        target_pos='',
        target_morph='',
        target_head=-1,
        target_deprel='',
        aligned_source_pos_list='',
        polysemy_count_source=0,
        polysemy_count_target=0,
    ):
        self.stacked_features = (
            stacked_features if stacked_features is not None else []
        )
        self.source_token_count = source_token_count  # Not used.
        self.target_token_count = target_token_count  # Not used.
        # Not used.
        self.source_target_token_count_ratio = source_target_token_count_ratio
        self.token = token
        self.left_context = left_context
        self.right_context = right_context
        self.first_aligned_token = first_aligned_token
        self.left_alignment = left_alignment
        self.right_alignment = right_alignment
        self.is_stopword = is_stopword  # Not used (at least for En-De).
        self.is_punctuation = is_punctuation
        self.is_proper_noun = is_proper_noun  # Not used (at least for En-De).
        self.is_digit = is_digit
        self.highest_order_ngram_left = highest_order_ngram_left
        self.highest_order_ngram_right = highest_order_ngram_right
        self.backoff_behavior_left = backoff_behavior_left  # Not used.
        self.backoff_behavior_middle = backoff_behavior_middle  # Not used.
        self.backoff_behavior_right = backoff_behavior_right  # Not used.
        self.source_highest_order_ngram_left = source_highest_order_ngram_left
        self.source_highest_order_ngram_right = source_highest_order_ngram_right
        self.pseudo_reference = pseudo_reference  # Not used in the WMT16+ data.
        self.target_pos = target_pos
        self.target_morph = target_morph  # Not used.
        self.target_head = target_head
        self.target_deprel = target_deprel
        self.aligned_source_pos_list = aligned_source_pos_list
        self.polysemy_count_source = polysemy_count_source  # Not used.
        self.polysemy_count_target = polysemy_count_target  # Not used.


class LinearWordQESentence:
    """Represents a sentence (word features and their labels)."""

    @staticmethod
    def create_stop_symbol():
        """Generates dummy features for a stop symbol."""
        return LinearWordQETokenFeatures(
            token='__STOP__',
            left_context='__STOP__',
            right_context='__STOP__',
            first_aligned_token='__STOP__',
            left_alignment='__STOP__',
            right_alignment='__STOP__',
            target_pos='__STOP__',
            aligned_source_pos_list='__STOP__',
            target_morph='__STOP__',
            target_deprel='__STOP__',
        )

    def __init__(self):
        self.sentence_word_features = []
        self.sentence_word_labels = []

    def num_words(self):
        """Returns the number of words of the sentence."""
        return len(self.sentence_word_features)

    def create_from_sentence_pair(
        self,
        source_words,
        target_words,
        alignments,
        source_pos_tags=None,
        target_pos_tags=None,
        target_parse_heads=None,
        target_parse_relations=None,
        target_ngram_left=None,
        target_ngram_right=None,
        target_stacked_features=None,
        labels=None,
    ):
        """Creates an instance from source/target token and alignment
        information."""
        self.sentence_word_features = []
        aligned_source_words = [[] for _ in target_words]
        for source, target in alignments:
            aligned_source_words[target].append(source)
        aligned_source_words = [
            sorted(aligned) for aligned in aligned_source_words
        ]
        if source_pos_tags is None:
            source_pos_tags = ['' for _ in source_words]
        if target_pos_tags is None:
            target_pos_tags = ['' for _ in target_words]
        if target_parse_heads is None:
            target_parse_heads = [-1 for _ in target_words]
        if target_parse_relations is None:
            target_parse_relations = ['' for _ in target_words]
        if target_ngram_left is None:
            target_ngram_left = [-1 for _ in target_words]
        if target_ngram_right is None:
            target_ngram_right = [-1 for _ in target_words]
        if target_stacked_features is None:
            target_stacked_features = ['' for _ in target_words]
        if labels is not None:
            if len(labels) != len(target_words):
                # WMT18 format with labels for the gaps.
                assert len(labels) == 2 * len(target_words) + 1
                labels = labels[1::2]
        for i in range(len(target_words)):
            word = target_words[i]
            tag = target_pos_tags[i]
            parse_head = int(target_parse_heads[i])  # TODO: don't cast here.
            parse_relation = target_parse_relations[i]
            ngram_left = int(target_ngram_left[i])  # TODO: don't cast here.
            ngram_right = int(target_ngram_right[i])  # TODO: don't cast here.
            if not target_stacked_features[i]:
                stacked_features = None
            else:
                stacked_features = [
                    float(p) for p in target_stacked_features[i].split('|')
                ]
            if i == 0:
                previous_word = '<s>'
                # previous_tag = '<s>'
            else:
                previous_word = target_words[i - 1]
                # previous_tag = target_pos_tags[i-1]
            if i == len(target_words) - 1:
                next_word = '</s>'
                # next_tag = '</s>'
            else:
                next_word = target_words[i + 1]
                # next_tag = target_pos_tags[i+1]
            if len(aligned_source_words[i]) == 0:
                source_word = '__unaligned__'
                previous_source_word = '__unaligned__'
                next_source_word = '__unaligned__'
                source_tag = '__unaligned__'
                # previous_source_tag = '__unaligned__'
                # next_source_tag = '__unaligned__'
            else:
                # Concatenate all source words in order of appearance.
                # The previous word is the one before the first aligned
                # source word; the next word is the one after the last
                # aligned word.
                source_word = '|'.join(
                    [source_words[j] for j in aligned_source_words[i]]
                )
                source_tag = '|'.join(
                    [source_pos_tags[j] for j in aligned_source_words[i]]
                )
                j = aligned_source_words[i][0]
                if j == 0:
                    previous_source_word = "<s>"
                    # previous_source_tag = "<s>"
                else:
                    previous_source_word = source_words[j - 1]
                    # previous_source_tag = source_pos_tags[j-1]
                if j == len(source_words) - 1:
                    next_source_word = "</s>"
                    # next_source_tag = "</s>"
                else:
                    next_source_word = source_words[j + 1]
                    # next_source_tag = source_pos_tags[j+1]

            word_features = LinearWordQETokenFeatures(
                stacked_features=stacked_features,
                source_token_count=len(source_words),
                target_token_count=len(target_words),
                source_target_token_count_ratio=float(len(source_words))
                / len(target_words),
                token=word,
                is_punctuation=all([c in string.punctuation for c in word]),
                is_digit=word.isdigit(),
                target_pos=tag,
                left_context=previous_word,
                right_context=next_word,
                first_aligned_token=source_word,
                aligned_source_pos_list=source_tag,
                left_alignment=previous_source_word,
                right_alignment=next_source_word,
                target_head=parse_head,
                target_deprel=parse_relation,
                highest_order_ngram_left=ngram_left,
                highest_order_ngram_right=ngram_right,
            )

            self.sentence_word_features.append(word_features)
            self.sentence_word_labels.append(
                labels[i] if labels is not None else ''
            )



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
import threading
import uuid

logger = logging.getLogger(__name__)


class TrackingLogger:
    class ActiveRun:
        def __init__(self, run_uuid, experiment_id):
            self.run_uuid = run_uuid
            self.experiment_name = experiment_id

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            return exc_type is None

    def __init__(self):
        self._experiment_id = None
        self._experiment_name = None
        self._active_run_uuids = []

    def configure(
        self, run_uuid, experiment_name,
        run_name=None, nest_run=True, *args, **kwargs
    ):
        if len(self._active_run_uuids) > 0 and not nest_run:
            raise Exception(
                (
                    "A run is already active. To start a nested run, call "
                    "start_nested_run(), or configure() with nest_run=True"
                )
            )
        if not self._active_run_uuids:
            self._experiment_name = experiment_name
            self._experiment_id = 0
            self._run_name = run_name
        if run_uuid is None:
            self._active_run_uuids.append(uuid.uuid4().hex)
        else:
            self._active_run_uuids.append(run_uuid)

        return TrackingLogger.ActiveRun(
            run_uuid=self._active_run_uuids[-1],
            experiment_id=self._experiment_id,
        )

    def start_nested_run(self, run_name=None):
        return self.configure(
            run_uuid=run_name, experiment_name=None, nest_run=True
        )

    @property
    def run_uuid(self):
        return self._active_run_uuids[-1] if self._active_run_uuids else None

    @property
    def experiment_id(self):
        return self._experiment_id

    @property
    def experiment_name(self):
        return self._experiment_name

    @property
    def run_name(self):
        return self._run_name

    def should_log_artifacts(self):
        return False

    def get_tracking_uri(self):
        return None

    @staticmethod
    def log_metric(key, value):
        pass

    @staticmethod
    def log_param(key, value):
        pass

    @staticmethod
    def log_artifact(local_path, artifact_path=None):
        pass

    @staticmethod
    def log_artifacts(local_dir, artifact_path=None):
        return None

    @staticmethod
    def get_artifact_uri():
        return None

    @staticmethod
    def end_run():
        pass


class MLflowLogger:
    def __init__(self):
        self.always_log_artifacts = False
        self._experiment_name = None
        self._run_name = None

    def configure(
        self,
        run_uuid,
        experiment_name,
        tracking_uri,
        run_name=None,
        always_log_artifacts=False,
        create_run=True,
        create_experiment=True,
        nest_run=True,
    ):
        if mlflow.active_run() and not nest_run:
            logger.info('Ending previous MLFlow run: {}.'.format(self.run_uuid))
            mlflow.end_run()

        self.always_log_artifacts = always_log_artifacts
        self._experiment_name = experiment_name
        self._run_name = run_name

        # MLflow specific
        if tracking_uri:
            mlflow.set_tracking_uri(tracking_uri)

        if run_uuid:
            existing_run = MlflowClient().get_run(run_uuid)
            if not existing_run and not create_run:
                raise FileNotFoundError(
                    'Run ID {} not found under {}'.format(
                        run_uuid, mlflow.get_tracking_uri()
                    )
                )

        experiment_id = self._retrieve_mlflow_experiment_id(
            experiment_name, create=create_experiment
        )
        return mlflow.start_run(
            run_uuid, experiment_id=experiment_id,
            run_name=run_name, nested=nest_run
        )

    def start_nested_run(self, run_name=None):
        return mlflow.start_run(run_name=run_name, nested=True)

    @property
    def run_uuid(self):
        return mlflow.tracking.fluent.active_run().info.run_uuid

    @property
    def experiment_id(self):
        return mlflow.tracking.fluent.active_run().info.experiment_id

    @property
    def experiment_name(self):
        # return MlflowClient().get_experiment(self.experiment_id).name
        return self._experiment_name

    def should_log_artifacts(self):
        return self.always_log_artifacts or self._is_remote()

    @staticmethod
    def get_tracking_uri():
        return mlflow.get_tracking_uri()

    @staticmethod
    def log_metric(key, value):
        mlflow.log_metric(key, value)

    @staticmethod
    def log_param(key, value):
        mlflow.log_param(key, value)

    @staticmethod
    def log_artifact(local_path, artifact_path=None):
        t = threading.Thread(
            target=mlflow.log_artifact,
            args=(local_path,),
            kwargs={'artifact_path': artifact_path},
            daemon=True,
        )
        t.start()

    @staticmethod
    def log_artifacts(local_dir, artifact_path=None):
        def send(dpath, e, path):
            mlflow.log_artifacts(dpath, artifact_path=path)
            e.set()

        event = threading.Event()
        t = threading.Thread(
            target=send, args=(local_dir, event, artifact_path), daemon=True
        )
        t.start()
        return event

    @staticmethod
    def get_artifact_uri():
        return mlflow.get_artifact_uri()

    @staticmethod
    def end_run():
        mlflow.end_run()

    def _is_remote(self):
        return not mlflow.tracking.utils._is_local_uri(
            mlflow.get_tracking_uri()
        )

    @staticmethod
    def _retrieve_mlflow_experiment_id(name, create=False):
        experiment_id = None
        if name:
            existing_experiment = MlflowClient().get_experiment_by_name(name)
            if existing_experiment:
                experiment_id = existing_experiment.experiment_id
            else:
                if create:
                    experiment_id = mlflow.create_experiment(name)
                else:
                    raise Exception(
                        'Experiment "{}" not found in {}'.format(
                            name, mlflow.get_tracking_uri()
                        )
                    )
        return experiment_id


try:
    import mlflow
    from mlflow.tracking import MlflowClient

    tracking_logger = MLflowLogger()
except ImportError:
    tracking_logger = TrackingLogger()



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import warnings
from collections import defaultdict

import torchtext

from kiwi.constants import PAD, START, STOP, UNALIGNED, UNK, UNK_ID


def _default_unk_index():
    return UNK_ID  # should be zero


class Vocabulary(torchtext.vocab.Vocab):
    """Defines a vocabulary object that will be used to numericalize a field.

    Attributes:
        freqs: A collections.Counter object holding the frequencies of tokens
            in the data used to build the Vocab.
        stoi: A collections.defaultdict instance mapping token strings to
            numerical identifiers.
        itos: A list of token strings indexed by their numerical identifiers.
    """

    def __init__(
        self,
        counter,
        max_size=None,
        min_freq=1,
        specials=None,
        vectors=None,
        unk_init=None,
        vectors_cache=None,
        rare_with_vectors=True,
        add_vectors_vocab=False,
    ):
        """Create a Vocab object from a collections.Counter.

        Arguments:
            counter: collections.Counter object holding the frequencies of
                each value found in the data.
            max_size: The maximum size of the vocabulary, or None for no
                maximum. Default: None.
            min_freq: The minimum frequency needed to include a token in the
                vocabulary. Values less than 1 will be set to 1. Default: 1.
            specials: The list of special tokens (e.g., padding or eos) that
                will be prepended to the vocabulary in addition to an <unk>
                token. Default: ['<pad>']
            vectors: One of either the available pretrained vectors
                or custom pretrained vectors (see Vocab.load_vectors);
                or a list of aforementioned vectors
            unk_init (callback): by default, initialize out-of-vocabulary word
                vectors to zero vectors; can be any function that takes in a
                Tensor and returns a Tensor of the same size.
                Default: torch.Tensor.zero_
            vectors_cache: directory for cached vectors.
                Default: '.vector_cache'
            rare_with_vectors: if True and a vectors object is passed, then
                it will add words that appears less than min_freq but are in
                vectors vocabulary. Default: True.
            add_vectors_vocab: by default, the vocabulary is built using only
                words from the provided datasets. If this flag is true, the
                vocabulary will add words that are not in the datasets but are
                in the vectors vocabulary (e.g. words from polyglot vectors).
                Default: False.
        """
        if specials is None:
            specials = ['<pad>']

        self.freqs = counter
        counter = counter.copy()
        min_freq = max(min_freq, 1)

        self.itos = list(specials)
        # frequencies of special tokens are not counted when building vocabulary
        # in frequency order
        for tok in specials:
            del counter[tok]

        max_size = None if max_size is None else max_size + len(self.itos)

        # sort by frequency, then alphabetically
        words_and_frequencies = sorted(counter.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)

        if not isinstance(vectors, list) and vectors is not None:
            vectors = [vectors]

        # add words that appears less than min_freq but are in embeddings
        # vocabulary
        for word, freq in words_and_frequencies:
            if freq < min_freq:
                if vectors is not None and rare_with_vectors:
                    for v in vectors:
                        if word in v.stoi:
                            self.itos.append(word)
                else:
                    break
            elif len(self.itos) == max_size:
                break
            else:
                self.itos.append(word)

        if add_vectors_vocab:
            if (
                max_size is not None
                and sum(v.stoi for v in vectors) + len(self.itos) > max_size
            ):
                warnings.warn(
                    'Adding the vectors vocabulary will make '
                    'len(vocab) > max_vocab_size!'
                )
            vset = set()
            for v in vectors:
                vset.update(v.stoi.keys())
            v_itos = vset - set(self.itos)
            self.itos.extend(list(v_itos))

        self.stoi = defaultdict(_default_unk_index)
        # stoi is simply a reverse dict for itos
        self.stoi.update({tok: i for i, tok in enumerate(self.itos)})

        self.vectors = None
        if vectors is not None:
            self.load_vectors(vectors, unk_init=unk_init, cache=vectors_cache)
        else:
            assert unk_init is None and vectors_cache is None


def merge_vocabularies(vocab_a, vocab_b, max_size=None, vectors=None, **kwargs):
    merged = vocab_a.freqs + vocab_b.freqs
    return Vocabulary(
        merged,
        specials=[UNK, PAD, START, STOP, UNALIGNED],
        max_size=max_size,
        vectors=vectors,
        **kwargs,
    )



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

# lowercased special tokens
UNK = '<unk>'
PAD = '<pad>'
START = '<bos>'
STOP = '<eos>'
UNALIGNED = '<unaligned>'

# special tokens id (don't edit this order)
# FIXME: avoid using these IDs since we don't really make sure they correspond
# to the above tokens
UNK_ID = 0
PAD_ID = 1
START_ID = 2
STOP_ID = 3
UNALIGNED_ID = 4

PAD_TAGS_ID = 2
# binary labels
OK = 'OK'
BAD = 'BAD'
OK_ID = 0
BAD_ID = 1
LABELS = [OK, BAD]

# this should be removed in the future
# fields
SOURCE = 'source'
TARGET = 'target'
PE = 'pe'
TARGET_TAGS = 'tags'
SOURCE_TAGS = 'source_tags'
GAP_TAGS = 'gap_tags'

TAGS = [TARGET_TAGS, SOURCE_TAGS, GAP_TAGS]

SENTENCE_SCORES = 'sentence_scores'
BINARY = 'binary'

TARGETS = [SENTENCE_SCORES, BINARY] + TAGS

ALIGNMENTS = 'alignments'
SOURCE_POS = 'source_pos'
TARGET_POS = 'target_pos'
TARGET_PARSE_HEADS = 'target_parse_heads'
TARGET_PARSE_RELATIONS = 'target_parse_relations'
TARGET_NGRAM_LEFT = 'target_ngram_left'
TARGET_NGRAM_RIGHT = 'target_ngram_right'
TARGET_STACKED = 'target_stacked'

# Constants for model output names
SENT_SIGMA = 'sentence_sigma'
LOSS = 'loss'
PREQEFV = 'PreQEFV'
POSTQEFV = 'PostQEFV'

# Standard Names for saving files
TRAIN = 'train'
DEV = 'dev'
TEST = 'test'
EVAL = 'eval'
VOCAB = 'vocab'
CONFIG = 'config'
STATE_DICT = 'state_dict'
VOCAB_FILE = 'vocab.torch'
MODEL_FILE = 'model.torch'
DATAFILE = 'dataset.torch'
OPTIMIZER = 'optim.torch'
BEST_MODEL_FILE = 'best_model.torch'
LAST_CHECKPOINT_FOLDER = 'latest_epoch'
TEMP_LAST_CHECKPOINT_FOLDER = 'temp_latest_epoch'
TRAINER = 'trainer.torch'
EPOCH = 'epoch'



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from distutils.util import strtobool

from kiwi.cli.better_argparse import ModelParser
from kiwi.cli.opts import PathType
from kiwi.models.quetch import QUETCH


def add_training_data_file_opts(parser):
    # Data options
    group = parser.add_argument_group('data')

    group.add_argument(
        '--train-source',
        type=PathType(exists=True),
        required=True,
        help='Path to training source file',
    )
    group.add_argument(
        '--train-target',
        type=PathType(exists=True),
        required=True,
        help='Path to training target file',
    )
    group.add_argument(
        '--train-alignments',
        type=str,
        required=True,
        help='Path to train alignments between source and target.',
    )
    group.add_argument(
        '--train-source-tags',
        type=PathType(exists=True),
        help='Path to training label file for source (WMT18 format)',
    )
    group.add_argument(
        '--train-target-tags',
        type=PathType(exists=True),
        help='Path to training label file for target',
    )

    group.add_argument(
        '--valid-source',
        type=PathType(exists=True),
        required=True,
        help='Path to validation source file',
    )
    group.add_argument(
        '--valid-target',
        type=PathType(exists=True),
        required=True,
        help='Path to validation target file',
    )
    group.add_argument(
        '--valid-alignments',
        type=str,
        required=True,
        help='Path to valid alignments between source and target.',
    )
    group.add_argument(
        '--valid-source-tags',
        type=PathType(exists=True),
        help='Path to validation label file for source (WMT18 format)',
    )
    group.add_argument(
        '--valid-target-tags',
        type=PathType(exists=True),
        help='Path to validation label file for target',
    )

    group.add_argument(
        '--valid-source-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for source',
    )
    group.add_argument(
        '--valid-target-pos',
        type=PathType(exists=True),
        help='Path to training PoS tags file for target',
    )

    return group


def add_predicting_data_file_opts(parser):
    # Data options
    group = parser.add_argument_group('data')

    group.add_argument(
        '--test-source',
        type=PathType(exists=True),
        required=True,
        help='Path to validation source file',
    )
    group.add_argument(
        '--test-target',
        type=PathType(exists=True),
        required=True,
        help='Path to validation target file',
    )
    group.add(
        '--test-alignments',
        type=PathType(exists=True),
        required=True,
        help='Path to test alignments between source and target.',
    )

    return group


def add_data_flags(parser):
    group = parser.add_argument_group('data processing options')
    group.add_argument(
        '--predict-target',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=True,
        help='Predict Target Tags. Leave unchanged for WMT17 format',
    )
    group.add_argument(
        '--predict-gaps',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict Gap Tags.',
    )
    group.add_argument(
        '--predict-source',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Predict Source Tags.',
    )
    group.add_argument(
        '--wmt18-format',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Read target tags in WMT18 format.',
    )

    group.add_argument(
        '--source-max-length',
        type=int,
        default=float("inf"),
        help='Maximum source sequence length',
    )
    group.add_argument(
        '--source-min-length',
        type=int,
        default=1,
        help='Truncate source sequence length.',
    )
    group.add_argument(
        '--target-max-length',
        type=int,
        default=float("inf"),
        help='Maximum target sequence length to keep.',
    )
    group.add_argument(
        '--target-min-length',
        type=int,
        default=1,
        help='Truncate target sequence length.',
    )

    return group


def add_vocabulary_opts(parser):
    group = parser.add_argument_group('vocabulary options')
    group.add_argument(
        '--source-vocab-size',
        type=int,
        default=None,
        help='Size of the source vocabulary.',
    )
    group.add_argument(
        '--target-vocab-size',
        type=int,
        default=None,
        help='Size of the target vocabulary.',
    )
    group.add_argument(
        '--source-vocab-min-frequency',
        type=int,
        default=1,
        help='Min word frequency for source vocabulary.',
    )
    group.add_argument(
        '--target-vocab-min-frequency',
        type=int,
        default=1,
        help='Min word frequency for target vocabulary.',
    )

    group.add_argument(
        '--keep-rare-words-with-embeddings',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Keep words that occur less then min-frequency but '
        'are in embeddings vocabulary.',
    )
    group.add_argument(
        '--add-embeddings-vocab',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Add words from embeddings vocabulary to source/target '
        'vocabulary.',
    )

    group.add_argument(
        '--embeddings-format',
        type=str,
        default='polyglot',
        choices=['polyglot', 'word2vec', 'fasttext', 'glove', 'text'],
        help='Word embeddings format. '
        'See README for specific formatting instructions.',
    )
    group.add_argument(
        '--embeddings-binary',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Load embeddings stored in binary.',
    )
    group.add_argument(
        '--source-embeddings',
        type=PathType(exists=True),
        help='Path to word embeddings file for source.',
    )
    group.add_argument(
        '--target-embeddings',
        type=PathType(exists=True),
        help='Path to word embeddings file for target.',
    )

    return group


def add_model_hyper_params_opts(training_parser):
    group = training_parser.add_argument_group('hyper-parameters')
    group.add_argument(
        '--bad-weight',
        type=float,
        default=3.0,
        help='Relative weight for bad labels.',
    )
    group.add_argument(
        '--window-size', type=int, default=3, help='Sliding window size.'
    )
    group.add_argument(
        '--max-aligned',
        type=int,
        default=5,
        help='Max number of alignments between source and target.',
    )
    group.add_argument(
        '--source-embeddings-size',
        type=int,
        default=50,
        help='Word embedding size for source.',
    )
    group.add_argument(
        '--target-embeddings-size',
        type=int,
        default=50,
        help='Word embedding size for target.',
    )
    group.add_argument(
        '--freeze-embeddings',
        type=lambda x: bool(strtobool(x)),
        nargs='?',
        const=True,
        default=False,
        help='Freeze embedding weights during training.',
    )
    group.add_argument(
        '--embeddings-dropout',
        type=float,
        default=0.0,
        help='Dropout rate for embedding layers.',
    )

    group.add_argument(
        '--hidden-sizes',
        type=int,
        nargs='+',
        default=[50],
        help='List of hidden sizes.',
    )
    group.add_argument(
        '--dropout',
        type=float,
        default=0.0,
        help='Dropout rate for linear layers.',
    )
    group.add_argument(
        '--init-type',
        type=str,
        default='uniform',
        choices=[
            'uniform',
            'normal',
            'constant',
            'glorot_uniform',
            'glorot_normal',
        ],
        help='Distribution type for parameters initialization.',
    )
    group.add_argument(
        '--init-support',
        type=float,
        default=0.1,
        help='Parameters are initialized over uniform distribution with '
        'support (-param_init, param_init). Use 0 to not use '
        'initialization.',
    )

    return group


def add_training_options(training_parser):
    add_training_data_file_opts(training_parser)
    add_data_flags(training_parser)
    add_vocabulary_opts(training_parser)
    add_model_hyper_params_opts(training_parser)


def add_predicting_options(predicting_parser):
    add_predicting_data_file_opts(predicting_parser)
    add_data_flags(predicting_parser)


def parser_for_pipeline(pipeline):
    if pipeline == 'train':
        return ModelParser(
            'quetch',
            'train',
            title=QUETCH.title,
            options_fn=add_training_options,
            api_module=QUETCH,
        )
    if pipeline == 'predict':
        return ModelParser(
            'quetch',
            'predict',
            title=QUETCH.title,
            options_fn=add_predicting_options,
            api_module=QUETCH,
        )

    return None



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import copy
import logging
from collections import defaultdict
from math import ceil
from pathlib import Path

import torch

from kiwi import constants as const
from kiwi.data.fieldsets.fieldset import Fieldset

logger = logging.getLogger(__name__)


def serialize_vocabs(vocabs, include_vectors=False):
    """Make vocab dictionary serializable.
    """
    serialized_vocabs = []

    for name, vocab in vocabs.items():
        vocab = copy.copy(vocab)
        vocab.stoi = dict(vocab.stoi)
        if not include_vectors:
            vocab.vectors = None
        serialized_vocabs.append((name, vocab))

    return serialized_vocabs


def deserialize_vocabs(vocabs):
    """Restore defaultdict lost in serialization.
    """
    vocabs = dict(vocabs)
    for name, vocab in vocabs.items():
        # Hack. Can't pickle defaultdict :(
        vocab.stoi = defaultdict(lambda: const.UNK_ID, vocab.stoi)
    return vocabs


def serialize_fields_to_vocabs(fields):
    """
    Save Vocab objects in Field objects to `vocab.pt` file.
    From OpenNMT
    """
    vocabs = fields_to_vocabs(fields)
    vocabs = serialize_vocabs(vocabs)
    return vocabs


def deserialize_fields_from_vocabs(fields, vocabs):
    """
    Load serialized vocabularies into their fields.
    """
    # TODO redundant deserialization
    vocabs = deserialize_vocabs(vocabs)
    return fields_from_vocabs(fields, vocabs)


def fields_from_vocabs(fields, vocabs):
    """
    Load Field objects from vocabs dict.
    From OpenNMT
    """
    vocabs = deserialize_vocabs(vocabs)
    for name, vocab in vocabs.items():
        if name not in fields:
            logger.debug(
                'No field "{}" for loading vocabulary; ignoring.'.format(name)
            )
        else:
            fields[name].vocab = vocab
    return fields


def fields_to_vocabs(fields):
    """
    Extract Vocab Dictionary from Fields Dictionary.
       Args:
          fields: A dict mapping field names to Field objects
       Returns:
          vocab: A dict mapping field names to Vocabularies
    """
    vocabs = {}
    for name, field in fields.items():
        if field is not None and 'vocab' in field.__dict__:
            vocabs[name] = field.vocab
    return vocabs


def save_vocabularies_from_fields(directory, fields, include_vectors=False):
    """
    Save Vocab objects in Field objects to `vocab.pt` file.
    From OpenNMT
    """
    vocabs = serialize_fields_to_vocabs(fields)
    vocab_path = Path(directory, const.VOCAB_FILE)
    torch.save({const.VOCAB: vocabs}, str(vocab_path))
    return vocab_path


def load_vocabularies_to_fields(vocab_path, fields):
    """Load serialized Vocabularies from disk into fields."""
    if Path(vocab_path).exists():
        vocabs_dict = torch.load(
            str(vocab_path), map_location=lambda storage, loc: storage
        )
        vocabs = vocabs_dict[const.VOCAB]
        fields = deserialize_fields_from_vocabs(fields, vocabs)
        logger.info('Loaded vocabularies from {}'.format(vocab_path))
        return all(
            [vocab_loaded_if_needed(field) for _, field in fields.items()]
        )
    return False


def load_vocabularies_to_datasets(vocab_path, *datasets):
    fields = {}
    for dataset in datasets:
        fields.update(dataset.fields)
    return load_vocabularies_to_fields(vocab_path, fields)


def vocab_loaded_if_needed(field):
    return not field.use_vocab or (hasattr(field, const.VOCAB) and field.vocab)


def save_vocabularies_from_datasets(directory, *datasets):
    fields = {}
    for dataset in datasets:
        fields.update(dataset.fields)
    return save_vocabularies_from_fields(directory, fields)


def build_vocabulary(fields_vocab_options, *datasets):
    fields = {}
    for dataset in datasets:
        fields.update(dataset.fields)

    for name, field in fields.items():
        if not vocab_loaded_if_needed(field):
            kwargs_vocab = fields_vocab_options[name]
            if 'vectors_fn' in kwargs_vocab:
                vectors_fn = kwargs_vocab['vectors_fn']
                kwargs_vocab['vectors'] = vectors_fn()
                del kwargs_vocab['vectors_fn']
            field.build_vocab(*datasets, **kwargs_vocab)


def load_datasets(directory, *datasets_names):
    dataset_path = Path(directory, const.DATAFILE)
    dataset_dict = torch.load(
        str(dataset_path), map_location=lambda storage, loc: storage
    )

    datasets = [dataset_dict[name] for name in datasets_names]
    return datasets


def save_datasets(directory, **named_datasets):
    """Pickle datasets to standard file in directory

    Note that fields cannot be saved as part of a dataset, so they are
    ignored.

    Args:
        directory (str or Path): directory where to save the datasets pickle.
        named_datasets (dict): mapping of name and respective dataset.
    """
    # Fields cannot be pickled
    # Saving field to a temporary list
    dataset_fields_tmp = []
    for dataset in named_datasets.values():
        dataset_fields_tmp.append(dataset.fields)
        dataset.fields = []

    logging.info('Saving preprocessed datasets...')
    dataset_path = Path(directory, const.DATAFILE)
    torch.save(named_datasets, str(dataset_path))

    # Reconstructing dataset.field from the temporary list
    for dataset, fields in zip(named_datasets.values(), dataset_fields_tmp):
        dataset.fields = fields


def save_training_datasets(directory, train_dataset, valid_dataset):
    ds_dict = {const.TRAIN: train_dataset, const.EVAL: valid_dataset}
    save_datasets(directory, **ds_dict)


def load_training_datasets(directory, fieldset):
    # FIXME: test if this works. Ideally, fields would be already contained
    # inside the loaded datasets.
    train_ds, valid_ds = load_datasets(directory, const.TRAIN, const.EVAL)

    # Remove fields not actually loaded (checking if they're required).
    fields = fieldset.fields
    for field in dict(fields):  # Make a copy so del can be used
        if not hasattr(train_ds.examples[0], field):
            for set_name in [Fieldset.TRAIN, Fieldset.VALID]:
                if fieldset.is_required(field, set_name):
                    raise AttributeError(
                        'Loaded {} dataset does not have a '
                        '{} field.'.format(set_name, field)
                    )
            del fields[field]

    train_ds.fields = fields
    valid_ds.fields = fields

    load_vocabularies_to_fields(
        Path(directory, const.VOCAB_FILE), fieldset.fields
    )

    return train_ds, valid_ds


def cross_split_dataset(dataset, splits):
    examples_per_split = ceil(len(dataset) / splits)
    for split in range(splits):
        held_out_start = examples_per_split * split
        held_out_stop = examples_per_split * (split + 1)

        held_out_examples = dataset[held_out_start:held_out_stop]
        held_in_examples = dataset[:held_out_start] + dataset[held_out_stop:]

        train_split = dataset.__class__(held_in_examples, dataset.fields)
        eval_split = dataset.__class__(held_out_examples, dataset.fields)

        yield train_split, eval_split


def save_file(file_path, data, token_sep=' ', example_sep='\n'):
    if data and isinstance(data[0], list):
        data = [token_sep.join(map(str, sentence)) for sentence in data]
    else:
        data = map(str, data)
    example_str = example_sep.join(data) + '\n'
    Path(file_path).write_text(example_str)


def save_predicted_probabilities(directory, predictions, prefix=''):
    directory = Path(directory)
    directory.mkdir(parents=True, exist_ok=True)
    for key, preds in predictions.items():
        if prefix:
            key = '{}.{}'.format(prefix, key)
        output_path = Path(directory, key)
        logger.info('Saving {} predictions to {}'.format(key, output_path))
        save_file(output_path, preds, token_sep=' ', example_sep='\n')


def read_file(path):
    """Reads a file into a list of lists of words.
    """
    with Path(path).open('r', encoding='utf8') as f:
        return [[token for token in line.strip().split()] for line in f]


def hter_to_binary(x):
    """Transform hter score into binary OK/BAD label.
    """
    return ceil(float(x))


def wmt18_to_target(batch, *args):
    """Extract target tags from wmt18 format file.
    """
    return batch[1::2]


def wmt18_to_gaps(batch, *args):
    """Extract gap tags from wmt18 format file.
    """
    return batch[::2]


def project(batch, *args):
    """Projection onto the first argument.

       Needed to create a postprocessing pipeline that implements the identity.
    """
    return batch


def filter_len(
    x,
    source_min_length=1,
    source_max_length=float('inf'),
    target_min_length=1,
    target_max_length=float('inf'),
):
    return (source_min_length <= len(x.source) <= source_max_length) and (
        target_min_length <= len(x.target) <= target_max_length
    )



"""This defines the class for defining sparse features in linear models."""

#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from .sparse_vector import SparseVector


class SparseFeatureVector(SparseVector):
    """A generic class for a sparse feature vector."""

    def __init__(
        self,
        save_to_cache=False,
        load_from_cache=False,
        cached_features_file=None,
    ):
        SparseVector.__init__(self)
        self.cached_features_file = cached_features_file
        self.save_to_cache = save_to_cache
        self.load_from_cache = load_from_cache

    def add_categorical_feature(self, name, value, allow_duplicates=False):
        """Add a categorical feature, represented internally as a binary
        feature."""
        fname = name + "=" + value
        assert allow_duplicates or fname not in self
        self[fname] = 1.0

    def add_binary_feature(self, name):
        """Add a binary feature."""
        if name in self:
            return
        self[name] = 1.0

    def add_numeric_feature(self, name, value):
        """Add a numeric feature."""
        self[name] = value

    def save_cached_features(self):
        """Save features to file."""
        self.cached_features_file.write(str(len(self)) + '\n')
        for key in self:
            self.cached_features_file.write(key + '\t' + str(self[key]) + '\n')

    def load_cached_features(self):
        """Load features from file."""
        num_features = int(self.cached_features_file.next())
        for i in range(num_features):
            key, value = (
                self.cached_features_file.next().rstrip('\n').split('\t')
            )
            self[key] = float(value)


class SparseBinaryFeatureVector(list):
    """A generic class for a sparse binary feature vector."""

    def __init__(
        self,
        feature_indices=None,
        save_to_cache=False,
        load_from_cache=False,
        cached_features_file=None,
    ):
        list.__init__(self)
        self.feature_indices = feature_indices
        self.cached_features_file = cached_features_file
        self.save_to_cache = save_to_cache
        self.load_from_cache = load_from_cache

    def add_categorical_feature(self, name, value):
        """Add a categorical feature, represented internally as a binary
        feature."""
        fname = name + "=" + value
        self.add_binary_feature(fname)

    def add_binary_feature(self, name):
        """Add a binary feature."""
        add = True
        index = self.feature_indices.get(name, -1)
        if index < 0:
            if not add:
                return
            else:
                index = self.feature_indices.add(name)
        self.append(index)

    def to_sparse_vector(self):
        """Convert to a SparseVector."""
        vector = SparseVector()
        for index in self:
            vector[index] = 1.0
        return vector

    def save_cached_features(self):
        """Save features to file."""
        self.cached_features_file.write(
            '\t'.join([str(key) for key in self]) + '\n'
        )

    def load_cached_features(self):
        """Load features from file."""
        self[:] = [
            int(key)
            for key in self.cached_features_file.next().rstrip('\n').split('\t')
        ]



#  OpenKiwi: Open-Source Machine Translation Quality Estimation
#  Copyright (C) 2019 Unbabel <openkiwi@unbabel.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from collections import Counter

from torchtext.data import Field


class SequenceLabelsField(Field):
    """Sequence of Labels.
    """

    def __init__(self, classes, *args, **kwargs):
        self.classes = classes
        self.vocab = None
        super().__init__(*args, **kwargs)

    def build_vocab(self, *args, **kwargs):
        specials = self.classes + [
            self.pad_token,
            self.init_token,
            self.eos_token,
        ]
        self.vocab = self.vocab_cls(Counter(), specials=specials, **kwargs)
